#define	VGAPAINT_CPOCTREE_C

#include "cpoctree.h"

static ulong nodecount;

///////////////////////////////////////////////////////////////////////
//
//	struct CTrnOctree *MakeCtrnOctree(struct FastPick *fp,uchar bits);
//
//	Creates a Color Translation Octree based on the 256-color palette
//	provided through the FastPick array.
//
//	The octree is calculated to the number of bits specified.  A value
//	of 8 computes the octree to full depth; intermediate values use
//	cube center tests to compute the final colors.  A value of 0 for
//	<bits> is NOT acceptable.
//
//	This code could probably be optimized significantly.

struct CTrnOctree *MakeCTrnOctree(struct FastPick *fp,uchar bits) {
	nodecount=0;
	bits = 6;
	return __MakeCTrnOctree(fp,256>>bits,255,0,0,0);
}

///////////////////////////////////////////////////////////////////////
//
//	struct CTrnOctree *__MakeCTrnOctree(struct FastPick *fp,uchar minval,uchar cval,uchar r,uchar g,uchar b);
//
//	Internal routine used to compute the CTrn octree.
//
//	Each color value is effectively surrounded by an infinite spheres.
//	Whenever two spheres intersect, they form a plane intersecting the
//	line segment connecting the two centers.  What this means is that
//	whenever all eight points of a cube are determined to be in the
//	same sphere, all points within that cube are also in that sphere.
//	That means that if we find all eight points are in the same sphere,
//	we can simply cut the tree there.
//
//	This code should probably be optimized to look for situations in
//	which the cube can be sliced into two rectangular solids, instead
//	of into eight halfcubes.

#pragma off (check_stack)		// please!

static struct CTrnOctree *__MakeCTrnOctree(struct FastPick *fp,uchar minval,uchar cval,uchar r,uchar g,uchar b) {

	struct CTrnOctree *cto;
	uchar c;

	if (cval>=31) printf("%02x %02x %02x (%02x) %ld\n",r,g,b,cval,nodecount);

	// test all eight points of the cube

	c =    FPick_Choose(fp,256,r     ,g     ,b     );
	if (c==FPick_Choose(fp,256,r+cval,g+cval,b+cval))
	if (c==FPick_Choose(fp,256,r+cval,g     ,b     ))
	if (c==FPick_Choose(fp,256,r     ,g+cval,b+cval))
	if (c==FPick_Choose(fp,256,r+cval,g+cval,b     ))
	if (c==FPick_Choose(fp,256,r     ,g     ,b+cval))
	if (c==FPick_Choose(fp,256,r     ,g+cval,b     ))
	if (c==FPick_Choose(fp,256,r+cval,g     ,b+cval))
	return (void *)c;

	// shrink the size of the cube

	if (cval<minval) {

		// can't shrink, already at largest size
		// return color value at midpoint of cube

		cval>>=1;
		return (void *)FPick_Choose(fp,256,r+cval,g+cval,b+cval);

	} else {
		// allocate an octree node

		if (!(cto = NEW(struct CTrnOctree))) {

			// no memory, return cube midpoint

			cval>>=1;
			return (void *)FPick_Choose(fp,256,r+cval,g+cval,b+cval);
		}

		// fill each halfcube

		cval>>=1;

		if (cval) {
			cto->path[0]=__MakeCTrnOctree(fp,minval,cval,r       ,g       ,b       );
			cto->path[1]=__MakeCTrnOctree(fp,minval,cval,r+cval+1,g       ,b       );
			cto->path[3]=__MakeCTrnOctree(fp,minval,cval,r+cval+1,g+cval+1,b       );
			cto->path[2]=__MakeCTrnOctree(fp,minval,cval,r       ,g+cval+1,b       );
			cto->path[6]=__MakeCTrnOctree(fp,minval,cval,r       ,g+cval+1,b+cval+1);
			cto->path[7]=__MakeCTrnOctree(fp,minval,cval,r+cval+1,g+cval+1,b+cval+1);
			cto->path[5]=__MakeCTrnOctree(fp,minval,cval,r+cval+1,g       ,b+cval+1);
			cto->path[4]=__MakeCTrnOctree(fp,minval,cval,r       ,g       ,b+cval+1);
		} else {
			cto->path[0]=(void *)FPick_Choose(fp,256,r  ,g  ,b  );
			cto->path[1]=(void *)FPick_Choose(fp,256,r+1,g  ,b  );
			cto->path[3]=(void *)FPick_Choose(fp,256,r+1,g+1,b  );
			cto->path[2]=(void *)FPick_Choose(fp,256,r  ,g+1,b  );
			cto->path[6]=(void *)FPick_Choose(fp,256,r  ,g+1,b+1);
			cto->path[7]=(void *)FPick_Choose(fp,256,r+1,g+1,b+1);
			cto->path[5]=(void *)FPick_Choose(fp,256,r+1,g  ,b+1);
			cto->path[4]=(void *)FPick_Choose(fp,256,r  ,g  ,b+1);
		};

		++nodecount;

		return cto;
	};
}


///////////////////////////////////////////////////////////////////
//
//	void	FreeCTrnOctree(struct CTrnOctree *ct);
//
//	Frees a CTrn octree.

void	FreeCTrnOctree(struct CTrnOctree *ct) {
	int i;
	struct CTrnOctree *p;

	for(i=0;i<8;i++) {
		p=ct->path[i];
		if (((ulong)p)>=256) FreeCTrnOctree(p); else free(p);
	}
}

