#include <ctype.h>
#include <io.h>
#include <stdio.h>

#include "file.h"
#include "vgapaint.h"
#include "list.h"
#include "listbox.h"
#include "request.h"
#include "io.h"

struct String {
	struct Node node;
	char attrib;
	char string[1];
};

void Rescan(void);
void AddFiles(char *);
struct String *filereq_getIndex(struct Listbox *,short);
void filereq_sldhandle(struct Slider *sld);
void filereq_display(struct Listbox *,struct String *,short,short);
void filereq_clear(struct Listbox *,short);
void filereq_scroll(struct Listbox *,short);
void filereq_clearWindow(struct Listbox *);

void fnstrcpy(char *,char *);
void fnaddpath(char *,char *);
void fnparent(char *);
boolean	fnmatch(char *extbuf,char *ext);

void StringTableSort(struct String **);

////////////////////////////////////////////////////

static char file_pathbuf[320]="";
static char file_namebuf[280]="";
static char namebuf[512]="";

static struct Field file_fields[]={
	{
		5,14,150,8,
		NULL,
		file_pathbuf,
		0,0,sizeof file_pathbuf-1,25,
		St0,St2
	},{
		6,127,78,8,
		NULL,
		file_namebuf,
		0,0,sizeof file_namebuf-1,13,
		St0,St2
	},
	{ -1 }
};

static struct Listbox file_lboxes[]={
	{
		6,26,72,96,12,12,NULL,
		0,0,-1,NULL,
		&filereq_getIndex,
		&filereq_display,
		&filereq_clear,
		&filereq_scroll,
		&filereq_clearWindow,
	},
	{ LB_END }
};

struct Rectangle file_pos_rects[]={
	{ 81, 39, 13, 70 },
	{ RECT_END }
};

struct Rectangle file_neg_rects[]={
	{  3, 12,154, 11 },
	{  4, 24, 76,100 },
	{  4,125, 82, 11 },
	{ 82, 40, 11, 68 },
	{ RECT_END }
};

struct Zone file_zones1[]={
	{  95, 24, 61, 11, ZN_3D+ZN_TEXTRENDER,NULL,"Parent" },
	{  95, 36, 61, 11, ZN_3D+ZN_TEXTRENDER,NULL,"Drives" },
	{  95, 48, 61, 11, ZN_3D+ZN_TEXTRENDER,NULL,"Assign" },
	{  96, 90, 59, 21, ZN_3D+ZN_TEXTRENDER,NULL,"Ok"	},
	{  96,112, 59, 21, ZN_3D+ZN_TEXTRENDER,NULL,"Cancel" },
	{ -1,-1,-1,-1 }
};

struct Zone file_zones2[]={
	{  81, 24, 13, 15,ZN_3D },
	{  81,109, 13, 15,ZN_3D },
	{ -1,-1,-1,-1 }
};

struct Zone file_zones3[]={
	{  95, 60, 61, 11,ZN_3D+ZN_INVERTED+ZN_TEXTRENDER,NULL,"    ",NULL },
	{ -1,-1,-1,-1 }
};

struct ZoneRange file_zr[]={
	{ &file_zr[1]	,file_zones1,NULL,-1,ZR_TAP },
	{ &file_zr[2]	,file_zones2,NULL,-1,ZR_TAP },
	{ NULL		,file_zones3,NULL,-1,ZR_TOGGLE },
};

struct Slider file_sliders[]={
	{	NULL,
		{ NULL },
		{ NULL },
		{ 83,41,9,66 },
		NULL,
		0,0,4,4,
		0,0,0,0,
		0,11,
		0,0,0,0,	/* positions */
		REQSLD_ONEHIT,
		&filereq_sldhandle
	}
};

struct Requester file_req={
	0,0,160,138,		/* rectangle */
	REQ_BORDER + REQ_POPCEN + REQ_FORCECOL,
	NULL,			/* title will be filled in */
	St2,			/* backfill */
	file_lboxes,
	file_fields,
	file_zr,
	file_sliders,
	file_pos_rects,
	file_neg_rects,
	NULL
};

static struct List fileList;
static struct String **fileTable;
static char **assignTable;

static char *file_ext;

void SetPrevName(char *name) {
	char c,*t;

	// scan the filename for a "\" or a ":"

	t=name;
	while(c=*t++)
		if (c=='\\' || c==':') {
			strcpy(file_pathbuf,name);
			fnparent(file_pathbuf);
			break;
		}

	fngetfn(file_namebuf,name);	
}

char *GetFileNameSave(char *title,char *ext) {
	char *n;

	n = GetFileName(title,ext);
	if (n) if (*n) {
		if (!access(n,F_OK))
			if (RequestBoolean("VP386 Warning","Yes",NULL,"No","File \"%s\" exists! Overwrite?",n)==2)
				return n;
			else
				return NULL;
	}
	return n;
}

static enum {
	FLIST_DIRECTORY,
	FLIST_DRIVES,
	FLIST_ASSIGNS
} filereq_mode;

char *GetFileName(char *title,char *ext) {
	boolean quit=FALSE,quit2=FALSE,okay=FALSE;
	mouse_coord x,y,b;
	ulong tick,tick2;

	int freq_height = 16;
	int freq_pheight;
	int freq_pwidth;

	/////////////////////////

	filereq_mode = FLIST_DIRECTORY;

	str_delete_spaces(file_namebuf);
	if (ext) {
		fn_change_extension(file_namebuf,ext);
		file_zones3[0].text[0]='.';
		strncpy(&file_zones3[0].text[1],ext,3);
		file_zones3[0].text[4]=0;
	} else {
		strcpy(&file_zones3[0].text[0],"*.*");
	};
	file_ext=ext;

	file_req.title=title;
	file_lboxes[0].seladd=NULL;
	file_lboxes[0].sel=-1;
	file_lboxes[0].alen=0;
	file_lboxes[0].textheight=sysfontheight;
	file_sliders[0].yclicks=1;

	// try the largest possible height for the file requester & reduce until we hit 8 or succeed

	freq_height = ((vidbm.height - 42)/sysfontheight)+1;

	if (opt.useLFN) file_lboxes[0].cwidth = 35; else file_lboxes[0].cwidth = 12;

	do {
		if (--freq_height < 8) return NULL;

		freq_pheight = freq_height*sysfontheight;
		freq_pwidth  = file_lboxes[0].cwidth * sysfontwidth;
		file_req.height			=  42 + freq_pheight;
		file_zones1[3].y		= - 6 + freq_pheight;
		file_zones1[4].y		=  16 + freq_pheight;
		file_zones2[1].y		=  13 + freq_pheight;
		file_pos_rects[0].height	= -26 + freq_pheight;
		file_neg_rects[3].height	= -28 + freq_pheight;
		file_neg_rects[1].height	=   4 + freq_pheight;
		file_neg_rects[2].y		=  29 + freq_pheight;
		file_sliders[0].boundbox.height = -30 + freq_pheight;
		file_fields[1].y		=  31 + freq_pheight;
		file_lboxes[0].cheight		= freq_height;
		file_lboxes[0].height		= freq_pheight;

		file_req.width			=  88 + freq_pwidth;
		file_pos_rects[0].x		=   9 + freq_pwidth;
		file_neg_rects[0].width		=  82 + freq_pwidth;
		file_neg_rects[1].width		=   4 + freq_pwidth;
		file_neg_rects[2].width		=   4 + freq_pwidth;
		file_neg_rects[3].x		=  10 + freq_pwidth;
		file_zones1[0].x		=  23 + freq_pwidth;
		file_zones1[1].x		=  23 + freq_pwidth;
		file_zones1[2].x		=  23 + freq_pwidth;
		file_zones1[3].x		=  24 + freq_pwidth;
		file_zones1[4].x		=  24 + freq_pwidth;
		file_zones2[0].x		=   9 + freq_pwidth;
		file_zones2[1].x		=   9 + freq_pwidth;
		file_zones3[0].x		=  23 + freq_pwidth;
		file_sliders[0].boundbox.x	=  11 + freq_pwidth;
		file_fields[0].width		=  78 + freq_pwidth;
		file_fields[1].width		=       freq_pwidth;
		file_lboxes[0].width		= freq_pwidth;
		file_lboxes[0].cwidth		= freq_pwidth/sysfontwidth;

		file_fields[0].cwidth		= file_fields[0].width/sysfontwidth;
		file_fields[1].cwidth		= file_fields[1].width/sysfontwidth;

		kbhit();

	} while (!OpenRequester(&file_req,0,0));

	mouse_hide();
	vd->bltpattern(&file_req.reqbm,arrowup  ,11 + freq_pwidth,26,9,11,ST0,ST2);
	vd->bltpattern(&file_req.reqbm,arrowdown,11 + freq_pwidth,15 + freq_pheight,9,11,ST0,ST2);
	vd->refresh();
	mouse_show();

	fileTable	= NULL;
	assignTable	= NULL;

	while (!quit2) {
		if (fileTable) {
			DisposeNodeTableCompletely((struct Node **)fileTable);
			fileTable=NULL;
		};
		if (assignTable) {
			free(assignTable);
			assignTable=NULL;
		}

		switch(filereq_mode) {
		case FLIST_DIRECTORY:
			Rescan();
			break;
		case FLIST_DRIVES:
			DriveScan();
			break;
		case FLIST_ASSIGNS:
			AssignScan();
			break;
		};

		file_sliders[0].yclicks=0;
		file_lboxes[0].alen=0;

		if (filereq_mode != FLIST_ASSIGNS) {
			if (!fileTable) break;

			StringTableSort(fileTable);

			file_sliders[0].yclicks=max(0,(long)fileTable[-1]-file_lboxes[0].cheight);
			file_lboxes[0].alen=(ulong)fileTable[-1];
		} else {
			file_sliders[0].yclicks = max(0,(long)assignTable[0]-file_lboxes[0].cheight);
			file_lboxes[0].alen = (long)assignTable[0];
		}
		file_sliders[0].cposy=file_sliders[0].pposy=0;
		file_lboxes[0].cpos=0;
		file_lboxes[0].sel=-1;
		file_lboxes[0].seladd=NULL;
		mouse_hide();
		DrawListbox(&file_lboxes[0]);
		SliderRecompute(&file_sliders[0]);
		mouse_show();
		quit=FALSE;
		while (!quit) {
			mouse_read2(&x,&y,&b);
			if (b) switch(CheckRequester(&file_req,x,y,b)) {
			case REQM_LISTBOX:
				mouse_wait_unkey();
				y=file_lboxes[0].sel;
				if (y<file_lboxes[0].alen && y>=0) {
					struct String *str=fileTable[y];

					switch(filereq_mode) {
					case FLIST_DRIVES:
						regs.h.ah=0x47;
						regs.h.dl=str->string[0]+1-'A';
						regs.x.esi=(ulong)&file_pathbuf[3];
						segread(&sregs);
						intdosx(&regs,&regs,&sregs);
						quit=TRUE;
						filereq_mode = FLIST_DIRECTORY;
						file_pathbuf[0]=str->string[0];
						file_pathbuf[1]=':';
						file_pathbuf[2]='\\';
						DisplayField(&file_fields[0]);
						break;
					case FLIST_DIRECTORY:
						if (str->attrib & FA_DIREC) {
							if (str->string[0]=='.' && str->string[1]=='.')
								fnparent(file_pathbuf);
							else	fnaddpath(file_pathbuf,str->string);

							DisplayField(&file_fields[0]);
							quit=TRUE;
						} else {
							strcpy(file_namebuf,str->string);
							DisplayField(&file_fields[1]);
						};
						break;
					case FLIST_ASSIGNS:
						strcpy(file_pathbuf,(char *)file_lboxes[0].seladd);
						strcat(file_pathbuf,":\\");
						DisplayField(&file_fields[0]);
						filereq_mode = FLIST_DIRECTORY;
						quit = TRUE;
						break;
					}
					break;
				};
				break;
			case REQM_FIELD+0:
				quit=TRUE;
				break;
			case REQM_FIELD+1:
				break;
			case RZoneMsg(1,0):
				mouse_hide();
				ZoneDraw(&file_zr[1],&file_zones2[0],0);
				mouse_show();
				SliderAbs(&file_sliders[0],0,file_lboxes[0].cpos-1);
				tick=read_tick()+11;
				do {
					tick2=read_tick();
					if (tick2>tick && (tick2-tick)>=1) {
						SliderAbs(&file_sliders[0],0,file_lboxes[0].cpos-1);
						tick=tick2;
					};
					mouse_read2(&x,&y,&b);
				} while(b);
				mouse_hide();
				ZoneErase(&file_zr[1],&file_zones2[0],0);
				vd->refresh();
				mouse_show();
				break;
			case RZoneMsg(1,1):
				mouse_hide();
				ZoneDraw(&file_zr[1],&file_zones2[1],0);
				mouse_show();
				SliderAbs(&file_sliders[0],0,file_lboxes[0].cpos+1);
				tick=read_tick()+11;
				do {
					tick2=read_tick();
					if (tick2>tick && (tick2-tick)>=1) {
						SliderAbs(&file_sliders[0],0,file_lboxes[0].cpos+1);
						tick=tick2;
					};
					mouse_read2(&x,&y,&b);
				} while(b);
				mouse_hide();
				ZoneErase(&file_zr[1],&file_zones2[1],0);
				vd->refresh();
				mouse_show();
				break;
			case RZoneMsg(0,0):
				fnparent(file_pathbuf);
				DisplayField(&file_fields[0]);
				quit=TRUE;
				break;
			case RZoneMsg(0,1):
				filereq_mode = FLIST_DRIVES;
				quit=TRUE;
				break;
			case RZoneMsg(0,2):
				filereq_mode = FLIST_ASSIGNS;
				quit = TRUE;
				break;
			case RZoneMsg(0,3):
				okay=TRUE;
				quit2=quit=TRUE;
				break;
			case RZoneMsg(0,4):
				quit2=quit=TRUE;
				break;
			case RZoneMsg(2,0):
				quit=TRUE;
				break;
			default: break;
			};
		};
	};

	if (fileTable) {
		DisposeNodeTableCompletely((struct Node **)fileTable);
		fileTable=NULL;
	};

	if (assignTable) free(assignTable);

	CloseRequester(&file_req);

	if (okay) {
		strcpy(namebuf,file_pathbuf);
		fnaddpath(namebuf,file_namebuf);

		return namebuf;
	} else return "";
}

void Rescan(void) {
	char path[144];

	strcpy(path,file_pathbuf);
	switch (path[strlen(path)-1]) {
	case ':': case '\\': break;
	default:
		strcat(path,"\\");
	};
	strcat(path,"*.*");

	NewList(&fileList);
	AddFiles(path);

	fileTable=(struct String **)MakeNodeTable(&fileList);
}

void	DriveScan(void) {
	short d;
	struct String *str;
	char tbuf1[4],tbuf2[37];
	union REGS regs;

	NewList(&fileList);

	for(d='A';d<='Z';d++) {
		sprintf(tbuf1,"%c:\\",d);
		regs.w.ax=0x2900;
		regs.x.esi=(unsigned int)tbuf1;
		regs.x.edi=(unsigned int)tbuf2;
		intdos(&regs,&regs);
		if (regs.h.al != 0xFF) {
			if (!(str=(struct String *)malloc(sizeof(struct String)))) return;
			str->attrib=0;
			str->string[0]=d;
			str->string[1]=':';
			str->string[2]=NULL;
			AddTail(&fileList,(struct Node *)str);
		};
	};

	fileTable=(struct String **)MakeNodeTable(&fileList);
}

void	AssignScan(void) {
	int	len;
	char	**p;
	struct VPAssign *vpa = NULL;

	len = VPAssignCount();

	if (!(assignTable = malloc(sizeof(char *)*(len+1))))
		return;

	p = assignTable;
	*p++ = (char *)len;

	while(vpa = VPAssignEnum(vpa)) *p++ = vpa->name;
}

void AddFiles(char *partspec) {
	struct find_t ffb;
	boolean error;
	struct String *str;
	char *name;
#if 0
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
#endif
	char pathtemp[256];

	VPAssignResolve(pathtemp,partspec);

#if 1

	// call VPfind* commands

	{
		struct VPFileFind vpff;
		ulong	handle;

		if (handle = VPfindfirst(pathtemp,_A_SUBDIR,&vpff)) {
			do {
				if (strcmp(vpff.lfn.lfname,".")) {
					if (!(vpff.lfn.attrib & FA_DIREC)) {
						if (file_ext && (file_zones3[0].flags & ZN_INVERTED)) {
							char *name = vpff.lfn.lfname;
							
							while(*name) ++name;
							while(name>vpff.lfn.lfname && *name!='.') --name;
//							if (stricmp(file_ext,name+1)) continue;	// reject filename
							if (!fnmatch(file_ext,name+1)) continue;
						}
					}

					if (strcmp(vpff.lfn.lfname,"..")) name=vpff.lfn.lfname;
					else name = ".. (Parent directory)";

					if (!(str = malloc(sizeof(struct String)+strlen(name))))
						break;	// out of memory, be sure to close find op

					strcpy(str->string,name);

					str->attrib = vpff.lfn.attrib;
					AddTail(&fileList,(struct Node *)str);
				}
			} while(!VPfindnext(handle,&vpff));
			VPfindclose(handle);
		}
	}
#else
	error=_dos_findfirst(pathtemp,_A_SUBDIR,&ffb);

	while(!error) {
		if (strcmp(name=ffb.name,".")) {
			if (!(ffb.attrib & FA_DIREC)) {
				if (file_ext && (file_zones3[0].flags & ZN_INVERTED)!=0) {
					ext[0]=ext[1]=0;
					_splitpath(name,NULL,NULL,fname,ext);

					// search the extensions list

					if (stricmp(file_ext,ext+1)) goto reject;
				};
			};
			if (!(str=malloc(sizeof(struct String)+12))) return;
			if (!strcmp(name,"..")) strcpy(str->string,".. (Parent)");
			else fnstrcpy(str->string,name);
			str->attrib=ffb.attrib;
			AddTail(&fileList,(struct Node *)str);
		};
reject:
		error=_dos_findnext(&ffb);
	};
#endif
}

void filereq_sldhandle(struct Slider *sld) {
	ScrollListbox(&file_lboxes[0],sld->cposy);
}

///////////////////////////////////

void filereq_display(struct Listbox *lb,struct String *entry,
	int dindex,int highlighted) {
	char buf[64];
	short y=lb->y+dindex*sysfontheight;

	if (!entry) {
		filereq_clear(lb,dindex);
		return;
	};

	if (filereq_mode == FLIST_ASSIGNS) {
		buf[63]=0;
		strncpy(buf,(char *)entry,63);
		strcat(buf,":");
		vd->bar(lb->destbm,lb->x,y,lb->width,sysfontheight,ST2);
		SysTextLim(lb->destbm,lb->x,y,buf,ST0,ST2,lb->cwidth);
	} else {

		buf[63]=0;
		strncpy(buf,entry->string,63);

		vd->bar(lb->destbm,lb->x,y,lb->width,sysfontheight,highlighted ? ST1 : ST2);
		SysTextLim(lb->destbm,lb->x,y,buf,entry->attrib & FA_DIREC ? ST3 : ST0,highlighted ? ST1 : ST2,lb->cwidth);
	}
	vd->refresh();
}

struct String *filereq_getIndex(struct Listbox *lb,int index) {
	if (index<0 || index>=lb->alen)
		return NULL;
	else {
		if (filereq_mode == FLIST_ASSIGNS)
			return (struct String *)assignTable[index+1];
		else
			return fileTable[index];
	}
}

void filereq_scroll(struct Listbox *lb,int dy) {
	mouse_hide();
	vd->vblit(lb->destbm,lb->x,lb->y,lb->width,lb->height,dy*sysfontheight);
	mouse_show();
}

void filereq_clear(struct Listbox *lb,int index) {
	mouse_hide();
	vd->bar(lb->destbm,lb->x,lb->y+sysfontheight*index,lb->width,sysfontheight,ST2);
	mouse_show();
}

void filereq_clearWindow(struct Listbox *lb) {
	mouse_hide();
	vd->bar(lb->destbm,lb->x,lb->y,lb->width,lb->height,ST2);
	mouse_show();
}

void fnstrcpy(char *y,char *x) {
	int z;
	char c;

	for (z=9;z;z--) {
		c=*x++;
		if (!c) goto abort;
		if (c=='.') break;
		*y++=c;
	};
	while(--z>0) *y++=' ';
	*y++='.';
	for (z=3;z;z--) {
		c=*x++;
		if (!c) break;
		*y++=c;
	};
abort:
	*y++=0;
}

void fnaddpath(char *path,char *add) {
	int i=0;
	char c;

	while(*path++) i++;

	path--;

	if (i) if (path[-1]!='\\' && path[-1]!=':')
		*path++='\\';

//	do {
//		if (!isspace(c=*add++)) *path++=c;
//	} while(c);

//	*path++=0;

	strcpy(path,add);
}

void fnparent(char *path) {
	int i=0;
	char c;

	while (*path++) ++i;
	if (!i) return;
	path--;
	while(i--) {
		c=*--path;
		if (c=='\\' || c==':') break;
	};
	if (i>0) if (path[-1] == ':' && *path=='\\') path++;
	*path=NULL;
}

void fngetfn(char *to,char *path) {
	int i=0;
	char c;

	while (*path++) ++i;
	if (!i) return;
	path--;
	while(i--) {
		c=*--path;
		if (c=='\\' || c==':') break;
	};
	if (i>0) while (*path == ':' || *path=='\\') path++;
	strcpy(to,path);
}

void	fn_change_extension(char *file,char *new_ext) {
	char drive[_MAX_DRIVE];
#if 0
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
#else
	char dir[256];
	char fname[256];
	char ext[16];
	char	newext[16],c,*t;
#endif

	t=newext;
	while(c=*new_ext++) if (c==';') break; else *t++=c;
	*t=0;

	_splitpath(file,drive,dir,fname,ext);
	_makepath(file,drive,dir,fname,newext);
}

boolean	fnmatch(char *extbuf,char *ext) {
	char *s;

	for(;;) {
		s=extbuf;
		while(*s && *s!=';') ++s;
		if (!strnicmp(extbuf,ext,s-extbuf)) return TRUE;
		if (!*s) break;
		extbuf = s+1;
	}
	return FALSE;
}

void	str_delete_spaces(char *s) {
	char c;
	char *t=s;

	do { c=*s++; if (c!=' ') *t++=c; } while(c);
}

#define	ss_swap(a,b) { swaptemp=st[a]; st[a]=st[b]; st[b]=swaptemp; }

void StringSort(struct String **st,long left,long right) {
	long i,last;
	struct String *swaptemp;

	if (left>=right) return;
	ss_swap(left,(left+right)>>1);
	last=left;
	for (i=left+1;i<=right;i++)
		switch ((st[i]->attrib & FA_DIREC) - (st[left]->attrib & FA_DIREC)) {
		case -FA_DIREC:	/* st[i] > st[left] */
			continue;
		case 0: 	/* st[i] ? st[left] */
			if (stricmp(st[i]->string,st[left]->string)>=0)
				continue;
			/* fall through */
		case FA_DIREC:	/* st[i] < st[left] */
			++last;
			ss_swap(last,i);
		};
	ss_swap(left,last);
	StringSort(st,left,last-1);
	StringSort(st,last+1,right);
}

void StringTableSort(struct String **st) {
	StringSort(st,1,(ulong)st[-1]-1);
}

void filereq_init(void) {

	/* read current path into path buffer */

	regs.h.ah=0x19;		/* dos read current drive */
	int386(0x21,&regs,&regs);

	sprintf(file_pathbuf,"%c:\\",'A'+regs.h.al);

	segread(&sregs);	/* prepare segment registers */
	regs.h.ah=0x47;		/* dos read drive path */
	regs.h.dl=0;
	regs.x.esi=FP_OFF((void far *)(file_pathbuf+3));
	sregs.ds=FP_SEG((void far *)(file_pathbuf+3));
	int386x(0x21,&regs,&regs,&sregs);
}


