#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "font.h"

void	vid13_RenderFont(BMP bm,struct Font *f,coord x1,coord y1,uchar xor1,uchar xor0,
		ulong flags,char *text,ulong maxchar) {

	uchar	and0;
	uchar	and1;
	long	data;
	uchar	*dataptr,*dptr;
	uchar	*vm,*vm1,*vm2;
	int	c,y,x,i;
	char	italic_offset;
	ulong	total_width=0;

	int	yskip,ysize;
	int	xskip=0,xsize;
	int	xskip2,xsize2;

	if (x1>bm->x2) return;			// too far right?

	yskip	= 0;		// first line of characters to draw
	ysize	= f->height;	// number of lines to draw

	if (y1<bm->y1) {
		yskip	-= (y1-bm->y1);
		ysize	+= (y1-bm->y1);
	}

	if (y1>bm->y2-(coord)f->height)
		ysize	-= (y1-(bm->y2-f->height));

	if (ysize <= 0) return;			// too far down or up

	if (x1<bm->x1)
		xskip=bm->x1-x1;
	xsize = bm->x2-x1-xskip;

	switch(flags & FRF_FGMODEMASK) {
	case FRF_FGXOR:
		and1 = 0xFF;
		break;
	case FRF_FGTRANS:
		and1 = 0xFF;
		xor1 = 0x00;
		break;
	default:	// FRF_FGNORMAL
		and1 = 0x00;
		break;
	}

	switch(flags & FRF_BGMODEMASK) {
	case FRF_BGXOR:
		and0 = 0xFF;
		break;
	case FRF_BGTRANS:
		and0 = 0xFF;
		xor0 = 0x00;
		break;
	default:	// FRF_FGNORMAL
		and0 = 0x00;
		break;
	}

	if (flags & FRF_REVERSE) {
		uchar t;

		t=and0; and0=and1; and1=t;
		t=xor0; xor0=xor1; xor1=t;
	}

	vm2 = bm->vm + (y1+yskip) * bm->bpr + x1;

	if (flags & FRF_ITALIC)
		vm2+=(f->height-1-yskip)>>1;

	while((c=(uchar)*text++) && xsize>0 && maxchar--) {
		fontKern	kern;
		fontWidth	width;
		fontSize	size;
		fontOffset	offset;
		uchar		last_carry;

		if (c<f->charlow || c>f->charhigh) c = f->charmissing;

		kern	= f->kernTable[c];
		width	= f->widthTable[c];
		size	= f->sizeTable[c];
		offset	= f->offsetTable[c];

		total_width += width;

		dataptr = f->data + (offset>>3) + yskip * f->dataBpr;

		vm1=vm2;	vm2 += size;
		italic_offset = ((f->baseline+yskip) & 1)^1;

		if (xskip>=width)
			xskip -= width;
		else {
			for(y=ysize;y;y--) {
				xskip2=xskip;
				xsize2=xsize;

				vm = vm1;	vm1 += bm->bpr;

				if (flags & FRF_ITALIC) {
					vm1 -= italic_offset;
					italic_offset ^= 1;
				}

				// clear data from left of cell to left of data

				if (kern>xskip2) {
					i=kern-xskip2;
					if (i>xsize2) i=xsize2;
					xsize2-=i;
					vm += xskip2;
					for(;i;i--) *vm++=(*vm & and0) ^ xor0;
					xskip2=0;
				} else {
					xskip2-=kern;
					vm += kern;
				}

				// obtain data from table - preshift as necessary

				x=width-1;

				if (x>xskip2) {
					x-=xskip2;
					if (x>xsize2) x=xsize2;
					xsize2 -= x;
					vm += xskip2;

					dptr = dataptr;		dataptr += f->dataBpr;
					data = *dptr++;

					if (flags & FRF_BOLD) {
						last_carry = (data & 0x01)<<7;
						data = (uchar)data | (((uchar)data) >> 1);
					};

					data += 0x00800000;
					data <<= (offset+xskip2) & 7;

					for(;x>0;x--) {
						if ((char)data < 0)	*vm++ = (*vm & and1) ^ xor1;
						else			*vm++ = (*vm & and0) ^ xor0;

						if (((signed long)(data<<=1))<0) {
							data = *dptr++;
							if (flags & FRF_BOLD) {
								data = (uchar)data | (((uchar)data) >> 1) | last_carry;
							}
							last_carry = (dptr[-1] & 0x01)<<7;
							data += 0x00800000;
						}
					}
					xskip2=0;
				} else {
					xskip2-=x;
					vm += x;
				}

				if (xskip2) { --xskip2; ++vm; } else if (xsize2) {
					--xsize2;
					if ((char)data < 0)	*vm++ = (*vm & and1) ^ xor1;
					else			*vm++ = (*vm & and0) ^ xor0;
				}

				// clear data from right of data to right of cell

				i=size-kern-width;
				if (i>xskip2) {
					i-=xskip2;
					if (i>xsize2) i=xsize2;
					xsize2-=i;
					vm += xskip2;
					for(;i;i--) *vm++=(*vm & and0) ^ xor0;
					xskip2=0;
				} else xskip2-=i;
			}
			if ((xskip -= width)<0) xskip=0;
			xsize -= width;
		}
	}

	vid_tap_rect(bm,x1,y1,total_width,f->height);
}

ulong	TextLength(struct Font *f,char *s) {
	ulong l=0;
	int c;

	while(c=(uchar)*s++) {
		if (c<f->charlow || c>f->charhigh) c = f->charmissing;
		l+=f->sizeTable[c];
	}

	return l;
}

// Format of font file:
//
// bytes 0-3	'VPFF'		VGAPaint 386 Font File
//
//	-- format 1 --
// bytes 4-7	0x00000000	Version 0 font (monospaced)
// bytes 8-11	width		Width of font
// bytes 12-15	height		Height of font
// bytes 16-19	baseline	Baseline of font
// byte  20	lowchar		Lowest character #
// byte  21	highchar	Highest character #
// byte  22-23	misschar	Missing character #
// byte  24-25	totalchar	Total # of characters defined
// bytes 26->	data		Raw data
//
//	-- format 2 --
// bytes 4-7	0x00000001	Version 1 font (proportional)
// bytes 8-11	maxwidth	Maximum width of font
// bytes 12-15	maxheight	Maximum height of font
// bytes 16-19	baseline	Baseline of font
// byte  20	lowchar		Lowest character #
// byte  21	highchar	Highest character #
// byte  22-23	misschar	Missing character #
// byte  24-25	totalchar	Total # of characters defined
// bytes 26-29	kernTable	Relative offset from start of font to kernTable
// bytes 30-33	widthTable	Relative offset from start of font to widthTable
// bytes 34-37	sizeTable	Relative offset from start of font to sizeTable
// bytes 38-41	offsetTable	Relative offset from start of font to offsetTable
// bytes 42-45	dataSize	# of bytes of font data to load
// bytes 46-49	dataTable	Relative offset from start of font to font data
// bytes 50-53	dataBpr		BytesPerRow of data table

struct Font *	LoadFont(char *fname) {
	FILE	*f;
	ulong	lbuf,format;
	ulong	offset=0;
	ulong	dataSize;
	struct Font *font;
	int	i;

	ulong	kernTblOff;
	ulong	widthTblOff;
	ulong	sizeTblOff;
	ulong	offTblOff;
	ulong	dataTblOff;

	if (!(f=VPfopen(fname,"rb"))) return NULL;

	if (!(font=malloc(sizeof(struct Font)))) {
		fclose(f);
		return NULL;
	}

	memset(font,0,sizeof(struct Font));

	if (1!=fread(&lbuf,4,1,f)) goto err;
	if (lbuf!='FFPV') goto err;

	if (1!=fread(&format,4,1,f)) goto err;

	switch(format) {
	case	0:	// version 0 (monospace) font
		if (1!=fread(&font->width,4,1,f)) goto err;
		if (1!=fread(&font->height,4,1,f)) goto err;
		if (1!=fread(&font->baseline,4,1,f)) goto err;
		font->charlow	 = fgetc(f);
		font->charhigh	 = fgetc(f);
		if (1!=fread(&font->charmissing,2,1,f)) goto err;
		if (1!=fread(&font->chartotal,2,1,f)) goto err;

		font->dataBpr = (font->width*font->chartotal+7)>>3;
		dataSize = font->dataBpr * font->height;

		break;
	case	1:	// version 1 (proportional) font
		if (1!=fread(&font->width,4,1,f)) goto err;
		if (1!=fread(&font->height,4,1,f)) goto err;
		if (1!=fread(&font->baseline,4,1,f)) goto err;
		font->charlow	 = fgetc(f);
		font->charhigh	 = fgetc(f);
		if (1!=fread(&font->charmissing,2,1,f)) goto err;
		if (1!=fread(&font->chartotal,2,1,f)) goto err;

		if (1!=fread(&kernTblOff,4,1,f)) goto err;
		if (1!=fread(&widthTblOff,4,1,f)) goto err;
		if (1!=fread(&sizeTblOff,4,1,f)) goto err;
		if (1!=fread(&offTblOff,4,1,f)) goto err;
		if (1!=fread(&dataSize,4,1,f)) goto err;
		if (1!=fread(&dataTblOff,4,1,f)) goto err;
		if (1!=fread(&font->dataBpr,4,1,f)) goto err;

		break;
	default:
		goto err;	// ???
	}

	// intermission! try to allocate tables needed

	if (!(font->kernTable	= malloc(sizeof(fontKern  )*font->chartotal))) goto err;
	if (!(font->widthTable	= malloc(sizeof(fontWidth )*font->chartotal))) goto abort0;
	if (!(font->sizeTable	= malloc(sizeof(fontSize  )*font->chartotal))) goto abort1;
	if (!(font->offsetTable	= malloc(sizeof(fontOffset)*font->chartotal))) goto abort2;
	if (!(font->data	= malloc(dataSize			   ))) goto abort3;

	switch(format) {
	case 0:
		if (1!=fread(font->data,dataSize,1,f)) goto abort4;

		memset(font->kernTable,0,sizeof(fontKern)*font->chartotal);
		for(i=0;i<font->chartotal;i++) {
			font->widthTable[i]=font->sizeTable[i]=font->width;
			font->offsetTable[i]=offset;
			offset += font->width;
		}
		break;
	case 1:
		if (fseek(f,kernTblOff,SEEK_SET)) goto abort4;
		if (1!=fread(font->kernTable,sizeof(fontKern)*font->chartotal,1,f)) goto abort4;

		if (fseek(f,widthTblOff,SEEK_SET)) goto abort4;
		if (1!=fread(font->widthTable,sizeof(fontWidth)*font->chartotal,1,f)) goto abort4;

		if (fseek(f,sizeTblOff,SEEK_SET)) goto abort4;
		if (1!=fread(font->sizeTable,sizeof(fontSize)*font->chartotal,1,f)) goto abort4;

		if (fseek(f,offTblOff,SEEK_SET)) goto abort4;
		if (1!=fread(font->offsetTable,sizeof(fontOffset)*font->chartotal,1,f)) goto abort4;

		if (fseek(f,dataTblOff,SEEK_SET)) goto abort4;
		if (1!=fread(font->data,dataSize,1,f)) goto abort4;
		break;
	}

	fclose(f);
	return font;

abort4:	free(font->data);
abort3:	free(font->offsetTable);
abort2:	free(font->sizeTable);
abort1:	free(font->widthTable);
abort0:	free(font->kernTable);
err:	fclose(f);
	free(font);
	return NULL;
}


void	FreeFont(struct Font *font) {
	free(font->data);
	free(font->offsetTable);
	free(font->sizeTable);
	free(font->widthTable);
	free(font->kernTable);
	free(font);
}

