#include <stdarg.h>

#include "vgapaint.h"
#include "listbox.h"
#include "video.h"

void DrawListbox(struct Listbox *lb) {
	DisplayListbox(lb,lb->cpos,0,lb->cheight);
}

void DisplayListbox(struct Listbox *lb,int index,int pos,int siz) {
	int i,j;

	if (pos>=lb->height || pos<0) return;
	if (index<0) index=0;
	if (siz<0) return;

	siz=min(siz,lb->cheight-pos);
	for(i=0;i<siz;++i) {
		j=index+i;
		if (j<lb->alen)
			(*lb->displayIndex)(lb,(void *)(*lb->getIndex)(lb,j),pos+i,(j==lb->sel));
		else
			(*lb->clearIndex)(lb,pos+i);
	};
	vd->refresh();
}

void SelectListbox(struct Listbox *lb,int index) {
	mouse_hide();
	if (lb->sel >= lb->cpos && lb->sel<(lb->cpos+lb->cheight))
		(*lb->displayIndex)(lb,lb->seladd,lb->sel-lb->cpos,FALSE);
	if (index>=0 && index<lb->alen) {
		lb->sel=index;
		lb->seladd=(void *)(*lb->getIndex)(lb,index);
		if (lb->sel >= lb->cpos && lb->sel<(lb->cpos+lb->height))
			(*lb->displayIndex)(lb,lb->seladd,lb->sel-lb->cpos,TRUE);
	} else {
		lb->sel=-1;
		lb->seladd=NULL;
	};
	vd->refresh();
	mouse_show();
}

void ScrollListbox(struct Listbox *lb,int index) {
	int dy,i;

	if (index>(lb->alen-lb->cheight)) index=lb->alen-lb->cheight;
	if (index<0) index=0;
	if (index == lb->cpos) return;

	dy=index-lb->cpos;
	lb->cpos=index;

	mouse_hide();
	if (abs(dy)>=lb->cheight) {
		(*lb->clearWindow)(lb);
		DrawListbox(lb);
	} else {
		(*lb->scrollIndex)(lb,-dy);
		if (dy<0) {
			i=0;
			while(dy++) {
				(*lb->displayIndex)(lb,(void *)(*lb->getIndex)(lb,index+i),i,(index+i == lb->sel));
				i++;
			};
		} else {
			i=lb->cheight;
			while(dy--) {
				--i;
				(*lb->displayIndex)(lb,(void *)(*lb->getIndex)(lb,index+i),i,(index+i == lb->sel));
			};
		};
	};
	vd->refresh();
	mouse_show();
}

void	NumIntoField(struct Field *fld,long num) {
	itoa(fld->lval = num,fld->buf,10);
	DisplayField(fld);
}

void DisplayField(struct Field *fld) {
	mouse_hide();
	vd->bar(fld->destbm,fld->x,fld->y,fld->width,fld->height,fld->back);
	if (fld->flags & FLD_SYSFONT)
		SysTextLim(fld->destbm,fld->x,fld->y,fld->buf+fld->cpos,fld->fore,fld->back,fld->cwidth);
	else
		vd->text6lim(fld->destbm,fld->x,fld->y,fld->buf+fld->cpos,fld->fore,fld->back,fld->cwidth);
	vd->refresh();
	mouse_show();
}

void DisplayFieldCursor(struct Field *fld,coord x) {
	int	cw,ch;

	if (fld->flags & FLD_SYSFONT) {
		cw = sysfontwidth;
		ch = sysfontheight;
	} else {
		cw = 6;
		ch = 8;
	}

	mouse_hide();
	vd->bar(fld->destbm,fld->x+cw*x,fld->y,cw,ch,0x80000000+(fld->fore ^ fld->back));
	mouse_show();
}

void ActivateField(struct Field *fld,coord x) {
	int key;
	coord cx;

	int	cw,ch;

	if (fld->flags & FLD_SYSFONT) {
		cw = sysfontwidth;
		ch = sysfontheight;
	} else {
		cw = 6;
		ch = 8;
	}

	x=(x-fld->x)/cw;
	if (x<0) x=0; else if (x>=fld->cwidth) x=fld->cwidth-1;
	x+=fld->cpos;
	if (x>(fld->flen=strlen(fld->buf))) x=fld->flen;

	mouse_wait_unkey();

	DisplayField(fld);
	DisplayFieldCursor(fld,x-fld->cpos);
	for(;;) {
		vd->refresh();
		if (mouse_buttons()) goto exit_hide;
		if ((key=kbd_check())!=-1) {
			mouse_hide();
			DisplayFieldCursor(fld,x-fld->cpos);
			switch(key) {
			case KBD_LEFT:
				if (x>0) --x;
				if (x<fld->cpos) {
					fld->cpos--;
					vd->hblit(fld->destbm,fld->x,fld->y,fld->width,fld->height,cw);
					vd->text6lim(fld->destbm,fld->x,fld->y,fld->buf+fld->cpos,fld->fore,fld->back,1);
				};
				break;
			case KBD_RIGHT:
				if (x<fld->flen) x++;
				if (x>=fld->cpos+fld->cwidth) {
					fld->cpos++;
					vd->hblit(fld->destbm,fld->x,fld->y,fld->width,fld->height,-cw);
					if (fld->cpos+fld->cwidth<=fld->flen)
						vd->text6lim(fld->destbm,fld->x+cw*(fld->cwidth-1),fld->y,fld->buf+fld->cpos+fld->cwidth-1,fld->fore,fld->back,1);
					else
						vd->bar(fld->destbm,fld->x+fld->width-cw,fld->y,cw,ch,fld->back);
				};
				break;
			case 0x153:	// delete
				if (x>=fld->flen) break;
				memmove(fld->buf+x,fld->buf+x+1,strlen(fld->buf+x+1));
				--fld->flen;
				cx=x-fld->cpos;
				if (cx<(fld->cwidth-1)) {
					vd->hblit(fld->destbm,fld->x+cw*cx,fld->y,fld->width-cw*cx,fld->height,-cw);
					if (fld->cpos+fld->cwidth<=fld->flen)
						vd->text6lim(fld->destbm,fld->x+cw*(fld->cwidth-1),fld->y,fld->buf+fld->cpos+fld->cwidth-1,fld->fore,fld->back,1);
					else
						vd->bar(fld->destbm,fld->x+fld->width-cw,fld->y,cw,ch,fld->back);
				};
				break;
			case 0x147:	// home
				x=fld->cpos=0;
				vd->bar(fld->destbm,fld->x,fld->y,fld->width,fld->height,fld->back);
				vd->text6lim(fld->destbm,fld->x,fld->y,fld->buf+fld->cpos,fld->fore,fld->back,fld->cwidth);
				break;
			case 0x14f:	// end
				fld->cpos=max(0,fld->flen+1-fld->cwidth);
				x=fld->flen;
				vd->bar(fld->destbm,fld->x,fld->y,fld->width,fld->height,fld->back);
				vd->text6lim(fld->destbm,fld->x,fld->y,fld->buf+fld->cpos,fld->fore,fld->back,fld->cwidth);
				break;
			case 0x1b:
			case '\r':
				goto exit;
			case 0x08:
				if (x>0) {
					memmove(fld->buf+x-1,fld->buf+x,strlen(fld->buf+x)+1);
					--x;
					fld->flen--;
					if (x<fld->cpos) fld->cpos--;
					else {
						vd->hblit(fld->destbm,fld->x+cw*(x-fld->cpos),fld->y,fld->width-cw*(x-fld->cpos),fld->height,-cw);
						if (fld->cpos+fld->cwidth<=fld->flen)
							vd->text6lim(fld->destbm,fld->x+cw*(fld->cwidth-1),fld->y,fld->buf+fld->cpos+fld->cwidth-1,fld->fore,fld->back,1);
						else
							vd->bar(fld->destbm,fld->x+fld->width-cw,fld->y,cw,ch,fld->back);
					};
				};

				break;
			case -1: break;
			default:
				if (key<0x100) {
					if (fld->flen>=fld->buflen) {
						vid_flash(FLD_FLASHTIME);
						break;
					};
					if (fld->flags & FLD_NUMERIC)
						if (!strchr("0123456789",key) && (x || key!='-')) {
							vid_flash(FLD_FLASHTIME);
							break;
						};
					memmove(fld->buf+x+1,fld->buf+x,strlen(fld->buf+x)+1);
					fld->buf[x]=key;
					fld->flen++;
					x++;
					if (x>=fld->cpos+fld->cwidth) {
						fld->cpos++;
						vd->hblit(fld->destbm,fld->x,fld->y,fld->width,fld->height,-cw);
					} else if (x<fld->flen) {
						vd->hblit(fld->destbm,fld->x+cw*(x-fld->cpos-1),fld->y,fld->width-cw*(x-fld->cpos-1),fld->height,cw);
					};
					vd->text6lim(fld->destbm,fld->x+cw*(x-fld->cpos-1),fld->y,fld->buf+x-1,fld->fore,fld->back,1);
				};
			};
			DisplayFieldCursor(fld,x-fld->cpos);
			mouse_show();
		};
	}
exit_hide:
	DisplayFieldCursor(fld,x-fld->cpos);
	mouse_hide();
exit:
	vd->refresh();
	mouse_show();	// we bypassed the regular show() when we used the goto

	fld->buf[fld->flen]=0;

	if (fld->flags & FLD_NUMERIC) {
		fld->lval=atol(fld->buf);
		if (fld->lval<fld->low || fld->lval>fld->high) {
			if (fld->lval<fld->low) fld->lval=fld->low;
			if (fld->lval>fld->high) fld->lval=fld->high;
			vid_flash(FLD_FLASHTIME);
			ltoa(fld->lval,fld->buf,10);
			DisplayField(fld);
		};
	};
	return;
}

/******** SLIDER ROUTINES ********/

/* SliderInit: Initialize a slider (duh) */

boolean SliderInit(struct Slider *sld) {
	int dir;

// if either SLD_ATTACHX or SLD_ATTACHY has been set, attach buttons
// onto slider

	if (sld->flags & SLD_ATTACHX) {
		if (sld->flags & SLD_OUTLINE3D) {
			sld->arrow1.x		= sld->boundbox.x - sld->arrow1.width - 2;
			sld->arrow2.x		= sld->boundbox.x + sld->boundbox.width + 2;
			sld->arrow1.y		= sld->arrow2.y		= sld->boundbox.y-2;
			sld->arrow1.height	= sld->arrow2.height	= sld->boundbox.height+4;
		} else {
			sld->arrow1.x		= sld->boundbox.x - sld->arrow1.width;
			sld->arrow2.x		= sld->boundbox.x + sld->boundbox.width;
			sld->arrow1.y		= sld->arrow2.y		= sld->boundbox.y;
			sld->arrow1.height	= sld->arrow2.height	= sld->boundbox.height;
		}
	} else if (sld->flags & SLD_ATTACHY) {
		if (sld->flags & SLD_OUTLINE3D) {
			sld->arrow1.x		= sld->arrow2.x		= sld->boundbox.x-2;
			sld->arrow1.y		= sld->boundbox.y - sld->arrow1.height-2;
			sld->arrow2.y		= sld->boundbox.y + sld->boundbox.height+2;
			sld->arrow1.width	= sld->arrow2.width	= sld->boundbox.width+4;
		} else {
			sld->arrow1.x		= sld->arrow2.x		= sld->boundbox.x;
			sld->arrow1.y		= sld->boundbox.y - sld->arrow1.height;
			sld->arrow2.y		= sld->boundbox.y + sld->boundbox.height;
			sld->arrow1.width	= sld->arrow2.width	= sld->boundbox.width;
		}
	}

/* see if we need to initialize a sprite */

	if (sld->knob) {	/* have a knob, so need to */
		uchar *save;
		if (!(save=malloc(sld->knob_width * sld->knob_height)))
			return FALSE;	/* failed */

		sld->knobspr.width=sld->knob_width;
		sld->knobspr.height=sld->knob_height;
		sprite_init(&sld->knobspr,sld->knob,save,sld->destbm,sld->knob_width);

		sld->xpixspc=sld->boundbox.width-1;
		sld->ypixspc=sld->boundbox.height-1;

	} else {		/* no autoknob - compute it's size */

		sld->knobspr.x=sld->boundbox.x;
		sld->knobspr.y=sld->boundbox.y;
		sld->knobspr.width=sld->boundbox.width;
		sld->knobspr.height=sld->boundbox.height;
	};

	SliderRecompute(sld);

	/* render slider's box if needed */

	if (sld->flags & SLD_OUTLINE3D) {
		vid_inv3d(sld->destbm, sld->boundbox.x-1,sld->boundbox.y-1,
					sld->boundbox.x+sld->boundbox.width,
					sld->boundbox.y+sld->boundbox.height);
		vid_draw3d(sld->destbm, sld->boundbox.x-2,sld->boundbox.y-2,
					sld->boundbox.x+sld->boundbox.width+1,
					sld->boundbox.y+sld->boundbox.height+1);
	};

	/* render arrows if we need to */

	if (dir=(sld->flags & SLD_ARROW1MASK))
		SliderDrawArrow(sld,&sld->arrow1,dir>>4);
	if (dir=(sld->flags & SLD_ARROW2MASK))
		SliderDrawArrow(sld,&sld->arrow2,dir>>8);

	vd->refresh();

	return TRUE;	/* success! */
}

// SliderRedraw:  Redraw areas

void SliderRedraw(struct Slider *sld) {
	int dir;

	if (sld->flags & SLD_OUTLINE3D) {
		vid_inv3d(sld->destbm, sld->boundbox.x-1,sld->boundbox.y-1,
					sld->boundbox.x+sld->boundbox.width,
					sld->boundbox.y+sld->boundbox.height);
		vid_draw3d(sld->destbm, sld->boundbox.x-2,sld->boundbox.y-2,
					sld->boundbox.x+sld->boundbox.width+1,
					sld->boundbox.y+sld->boundbox.height+1);
	};

	/* render arrows if we need to */

	if (dir=(sld->flags & SLD_ARROW1MASK))
		SliderDrawArrow(sld,&sld->arrow1,dir>>4);
	if (dir=(sld->flags & SLD_ARROW2MASK))
		SliderDrawArrow(sld,&sld->arrow2,dir>>8);

}

/* SliderDrawArrow: Render arrows if needed */

void SliderDrawArrow(struct Slider *sld,struct Rectangle *r,short dir) {
	coord x1,y1,x2,y2,x3,y3;

	switch(dir) {
	case 1:	x1=x2=y2=y3=2;
		y1=   r->height-3;
		x3=   r->width-3;
		break;
	case 2:	x1=y2=2;
		y1=y3=r->height-3;
		x2=   (r->width >> 1);
		x3=   r->width-3;
		break;
	case 3:	x1=y1=y2=2;
		x2=x3=r->width-3;
		y3=   r->height-3;
		break;
	case 4:	x1=x3=r->width-3;
		y1=   r->height-3;
		x2=y3=2;
		y2=   (r->height>>1);
		break;
	case 6:	x1=y1=x3=2;
		x2=r->width-3;
		y2=(r->height>>1);
		y3=r->height-3;
		break;
	case 7:	x1=y1=x2=2;
		y2=y3=r->height-3;
		x3=r->width-3;
		break;
	case 8:	x1=y1=y3=2;
		x2=(r->width>>1);
		y2=r->height-3;
		x3=r->width-3;
		break;
	case 9:	x1=y3=2;
		y1=y2=r->height-3;
		x2=x3=r->width-3;
		break;
	}
	x1+=r->x;
	x2+=r->x;
	x3+=r->x;
	y1+=r->y;
	y2+=r->y;
	y3+=r->y;
	vd->line(sld->destbm,x1,y1,x2,y2,ST0);
	vd->line(sld->destbm,x3,y3,x2,y2,ST0);
	vid_draw3d(sld->destbm,r->x,r->y,r->x+r->width-1,r->y+r->height-1);
}

/* SliderRecompute: Recompute a slider with changed click settings */

void SliderRecompute(struct Slider *sld) {
	/* if we are controlling a listbox, reset 'y' setting */

	if (sld->flags & SLD_LISTBOX)
		sld->yclicks=max(0,sld->lboxctrl->alen-sld->lboxctrl->cheight);

	if (sld->xclicks<0) sld->xclicks=0;
	if (sld->yclicks<0) sld->yclicks=0;
	if (!sld->knob) {
		sld->knob_width =max(sld->min_width ,sld->boundbox.width  / (sld->xclicks+1));
		sld->knob_height=max(sld->min_height,sld->boundbox.height / (sld->yclicks+1));
		sld->xpixspc=sld->boundbox.width-sld->knob_width;
		sld->ypixspc=sld->boundbox.height-sld->knob_height;
		sld->knobspr.x=sld->boundbox.x;		// knob size has changed, do this to prevent rendering outside box
		sld->knobspr.y=sld->boundbox.y;
		vd->bar(sld->destbm,sld->boundbox.x,sld->boundbox.y,sld->boundbox.width,sld->boundbox.height,ST2);
	};
	SliderAbs(sld,sld->cposx,sld->cposy);
}

/* SliderDraw: Draw a slider at it's current position */

void SliderDraw(struct Slider *sld) {
	long x,y;

	mouse_hide();
	if (sld->knob) {
		sprite_quikmove2(&sld->knobspr,sld->boundbox.x+sld->pposx-sld->min_width,sld->boundbox.y+sld->pposy-sld->min_height);
	} else {
//		vd->bar(sld->destbm,sld->boundbox.x,sld->boundbox.y,sld->boundbox.width,sld->boundbox.height,ST2);
		vd->bar(sld->destbm,
				sld->knobspr.x,
				sld->knobspr.y,
				sld->knob_width,
				sld->knob_height,
				ST2);
		vid_draw3d(sld->destbm,
			x = sld->knobspr.x = sld->boundbox.x+sld->pposx,
			y = sld->knobspr.y = sld->boundbox.y+sld->pposy,
			sld->boundbox.x+sld->pposx+sld->knob_width-1,
			sld->boundbox.y+sld->pposy+sld->knob_height-1);
		vd->bar(sld->destbm,x+1,y+1,sld->knob_width-2,sld->knob_height-2,ST2);
		sld->knobspr.width=sld->knob_width;
		sld->knobspr.height=sld->knob_height;
	};
	mouse_show();
	if (sld->updater) sld->updater(sld);
	if (sld->flags & SLD_LISTBOX)
		ScrollListbox(sld->lboxctrl,sld->cposy);

	/* all slider functions except SliderInit() go through this */

	vd->refresh();
}

/* SliderMove: Move a slider to a particular pixel position */

void SliderMove(struct Slider *sld,long x,long y) {
	if (x<0) x=0; else if (x>sld->xpixspc) x=sld->xpixspc;
	if (y<0) y=0; else if (y>sld->ypixspc) y=sld->ypixspc;

	sld->pposx=x;
	sld->pposy=y;
	if (!x)	sld->cposx=0;
	else	sld->cposx=scale32i(x,sld->xclicks,sld->xpixspc);
	if (!y) sld->cposy=0;
	else	sld->cposy=scale32i(y,sld->yclicks,sld->ypixspc);

	SliderDraw(sld);
}

/* SliderAbs: Move a slider to a particular click position */

void SliderAbs(struct Slider *sld,long x,long y) {
	if (x<0) x=0; else if (x>sld->xclicks) x=sld->xclicks;
	if (y<0) y=0; else if (y>sld->yclicks) y=sld->yclicks;

	sld->cposx=x;
	sld->cposy=y;
	if (!x)	sld->pposx=0;
	else	sld->pposx=scale32i(x,sld->xpixspc,sld->xclicks);
	if (!y)	sld->pposy=0;
	else	sld->pposy=scale32i(y,sld->ypixspc,sld->yclicks);

	SliderDraw(sld);
}

/* SliderDrag: Drag a slider using the mouse pointer */

void SliderDrag(struct Slider *sld) {
	mouse_coord x,y,b;
	long offx,offy,ox,oy;

	mouse_read2(&x,&y,&b);
	offx=x-sld->pposx;
	offy=y-sld->pposy;
	x=sld->pposx;
	y=sld->pposy;

	do {
		ox=x;
		oy=y;
		do {
			mouse_read2(&x,&y,&b);
			x-=offx;
			y-=offy;
			if (x<0) x=0; else if (x>sld->xpixspc) x=sld->xpixspc;
			if (y<0) y=0; else if (y>sld->ypixspc) y=sld->ypixspc;
		} while((b&1) && ox==x && oy==y);
		SliderMove(sld,x,y);
	} while(b & 1);
}

/* SliderClose: Dispose of a slider */

void SliderClose(struct Slider *sld) {

	/* if we have a knob, we must dispose of the sprite */

	if (sld->knob) {
		sprite_hide(&sld->knobspr);
		free(sld->knobspr.savebm.vm);
	};

}

/* SliderHandle: Deal with clicks on a slider */

ulong SliderHandle(struct Slider *sld,long x,long y) {

	long relx=x-sld->boundbox.x;
	long rely=y-sld->boundbox.y;
	long tx,ty;
	ulong id;

	/* check for a hit on arrows */

	if (sld->flags & SLD_ARROW1MASK)
		if (rectInside(&sld->arrow1,x,y)) {
			SliderAbs(sld,sld->cposx+sld->arrow1tapx,sld->cposy+sld->arrow1tapy);
			if (sld->flags & SLD_ARROW1HIT)
				mouse_wait_unkey();
			return 16;
		};
	if (sld->flags & SLD_ARROW2MASK)
		if (rectInside(&sld->arrow2,x,y)) {
			SliderAbs(sld,sld->cposx+sld->arrow2tapx,sld->cposy+sld->arrow2tapy);
			if (sld->flags & SLD_ARROW2HIT)
				mouse_wait_unkey();
			return 17;
		};

	tx=relx-sld->pposx;
	ty=rely-sld->pposy;

	/* check for a hit on the knob */

	if (sld->knob) {
		if (sprite_pointcoll(&sld->knobspr,x,y)) {
			SliderDrag(sld);
			return SLD_HITID+5;
		};
		tx-=sld->min_width;
		ty-=sld->min_height;
	} else {
		if (tx>=0)
			if (ty>=0)
				if (tx<sld->knob_width)
					if (ty<sld->knob_height) {
						SliderDrag(sld);
						return SLD_HITID+5;
					};
	};

	/* check for a hit in the bounding zone */

	if (relx<0) return 0;	/* it's not even in the bounding box */
	if (rely<0) return 0;
	if (relx>=sld->boundbox.width) return 0;
	if (rely>=sld->boundbox.height) return 0;

	id=SLD_HITID+5;

	if (tx<0) {
		tx=-sld->xbump;
		id--;
	} else if (tx>=sld->knob_width) {
		tx=sld->xbump;
		id++;
	} else {
		tx=0;
	};
	if (ty<0) {
		ty=-sld->ybump;
		id-=3;
	} else if (ty>=sld->knob_height) {
		ty=sld->ybump;
		id+=3;
	} else {
		ty=0;
	};

	SliderAbs(sld,sld->cposx+tx,sld->cposy+ty);

	return id;
}


/******** Miscellaneous functions ********/


/* call this function before OpenRequester() if dynamic textbox */

void	InitTextbox(struct Textbox *tb,struct Listbox *lb) {
	NewList(&tb->text);
	tb->lb=lb;
}

/* TextboxAdd: Add a string at end of textbox */

boolean TextboxAdd(struct Textbox *tb,char *string) {
	char *str;
	struct TextboxNode *tbn;

	if (!(str=malloc(strlen(string)+1))) return TRUE;
	if (!(tbn=malloc(sizeof(struct TextboxNode)))) {
		free(str);
		return TRUE;
	};
	strcpy(str,string);
	tbn->string=str;
	AddTail(&tb->text,(struct Node *)tbn);
	return FALSE;
}

/* TextboxPrintf: Formatted add to end of textbox */

boolean	TextboxPrintf(struct Textbox *tb,char *format, ...) {
	va_list val;
	boolean fail;

	va_start(val,format);
	vsprintf(packmem,format,val);
	fail=TextboxAdd(tb,packmem);
	va_end(val);
	return fail;
}

/* TextboxGlue: Glue all strings together into a table */

char **TextboxGlue(struct Textbox *tb) {
	struct Node *node,*node2;
	char **table,**temp;
	ulong siz=ListLength(&tb->text);

	if (!(table=(char **)malloc(sizeof(char *)*(siz+1))))
		return NULL;

	node=tb->text.prev;

	*table=(char *)siz;

	temp=table+1;

	while (siz--) {
		*temp++=((struct TextboxNode *)node)->string;
		node2=node->prev;
		free(node);
		node=node2;
	};

	tb->table=table;

	return table+1;
}

/* TextboxDissolve: Free a dynamic text list */

void	TextboxDissolve(struct Textbox *tb) {
	ulong siz=(ulong)tb->table[0];
	char **tbl=&tb->table[1];

	while(siz--)
		free(*tbl++);
	free(tb->table);
}

/* TextboxCreate: Create a Listbox from a Textbox */

void	TextboxCreate(struct Textbox *tb,boolean sysfont) {
	struct Listbox *lb;
	char **tbl;
	ulong leng;

	lb=tb->lb;

	if (-1==(lb->alen=(int)tb->table[0])) {
		leng=0;
		tbl=tb->table+1;
		while(*tbl++) ++leng;
		lb->alen=leng;
	};

	lb->getIndex=&TextboxGetIndex;
	if (sysfont)	lb->displayIndex=&TextboxDisplayIndex2;
	else		lb->displayIndex=&TextboxDisplayIndex;
	lb->clearIndex=&TextboxClearIndex;
	lb->scrollIndex=&TextboxScrollIndex;
	lb->clearWindow=&TextboxClearWindow;
	lb->listadd=(void *)&tb->table[1];
	lb->textheight = sysfont ? sysfontheight : 8;
	lb->width=lb->cwidth * (sysfont ? sysfontwidth : 6);
	lb->height=lb->cheight * (sysfont ? sysfontheight : 8);
	lb->cpos=0;
	lb->sel=-1;
	lb->seladd=NULL;

	/* only one not filled in is destbm */

}

void *	TextboxGetIndex(struct Listbox *lb,int index) {
	return (void *)(((char **)(lb->listadd))[index]);
}

void	TextboxDisplayIndex(struct Listbox *lb,void *entry,int dindex,int highlight) {
	color	backg;
	coord	y;

	y=lb->y+lb->textheight*dindex;
	backg=highlight ? ST1 : ST2;

	vd->bar(lb->destbm,lb->x,y,lb->width,lb->textheight,backg);
	if (entry)
		vd->text6lim(lb->destbm,lb->x,y,(char *)entry,highlight ? ST3 : ST0,backg,lb->cwidth);
}

void	TextboxDisplayIndex2(struct Listbox *lb,void *entry,int dindex,int highlight) {
	color	backg;
	coord	y;

	y=lb->y+lb->textheight*dindex;
	backg=highlight ? ST1 : ST2;

	vd->bar(lb->destbm,lb->x,y,lb->width,lb->textheight,backg);
	if (entry)
		SysTextLim(lb->destbm,lb->x,y,(char *)entry,highlight ? ST3 : ST0,backg,lb->cwidth);
}

void	TextboxClearIndex(struct Listbox *lb,int dindex) {
	vd->bar(lb->destbm,lb->x,lb->y+lb->textheight*dindex,lb->width,lb->textheight,ST2);
}

void	TextboxScrollIndex(struct Listbox *lb,int dy) {
	vd->vblit(lb->destbm,lb->x,lb->y,lb->width,lb->height,dy*lb->textheight);
}

void	TextboxClearWindow(struct Listbox *lb) {
	vd->bar(lb->destbm,lb->x,lb->y,lb->width,lb->height,ST2);
}

