#define PARM_C

#include <fcntl.h>

#include "vgapaint.h"
#include "inifile.h"
#include "io.h"
#include "parm.h"

#include "startup.con"

extern struct IniItem vp_ini_options[];

long	ppgetvalue(char *);
void	ppinstructions(void);
void	ppdefaults(void);

extern	char version_str[];

static int	ppargc;
static char **	ppargv;

boolean easter_decode = FALSE;

void ParamParse(int argc,char **argv,char *inistring) {
	char c,*s;
	boolean nohandlers=FALSE;

	program_name = argv[0];

	picbm->width=picbm->height=0;

	if (!argc) goto xit;

	ppargc=argc;
	ppargv=argv+1;
	while(ppargc--) {
		s=ppargc ? *ppargv++ : inistring;
		if (!s) continue;

		while(c=*s++) {
			switch(c) {
			case '-': case '/':
				switch(*s++) {
				case 'w':	/* width */
					picbm->width=max(16,ppgetvalue(s));
					break;
				case 'h':	/* height */
					picbm->height=max(16,ppgetvalue(s));
					break;
				case 'm':	// mouse pointer problems
					switch(*s++) {
					case 'b':	// bias
						ms_bias=0;
						break;
					case 'p':	// no pan
						ms_pan = 0;
						break;
					default:
						ppinstructions();
					};
					break;
				case 'd':	// debugging
					switch(*s++) {
					case 'L':	// logging - already handled in Init();
						break;
					case 'e':
						nohandlers=TRUE;
						break;
					case 'l':
						opt.useLFN = FALSE;
						break;
					case 'w':
						opt.wimp_mode = TRUE;
						break;
					case 'V':
						printf("Virus checking code disabled - AHL 7/3/98\n"); getch();
						break;
					case 'v':
						switch(*s++) {
						case '2':
							opt.noVESA20 = TRUE;
							break;
						case '3':
							opt.noVESA32bit = TRUE;
							break;
						case 'l':
							opt.noVESA_LFB = TRUE;
							break;
						case 'm':
							opt.noVESAmodesearch = TRUE;
							break;
						default:
							--s;
							opt.noVESA = TRUE;
						}
						break;
					case 'p':
						opt.nodirect_pal= TRUE;
						break;
					case 'r':
						opt.releaseTest = TRUE;
						break;
					default:
						ppinstructions();
					};
					break;
				case '~':
					easter_decode=TRUE;
					break;
				default:
					ppinstructions();
				};
				break;
			case '?':
				ppinstructions();
				break;
			}
		}
	};

	if (opt.wimp_mode) {
		opt.noHack		= TRUE;
		opt.stdVGAonly	= TRUE;
		opt.nodirect_pal= TRUE;
		opt.noVESA		= TRUE;
		opt.noVESA20	= TRUE;
		opt.noVESA32bit = TRUE;
		opt.noVESA_LFB	= TRUE;
		opt.noVESAmodesearch = TRUE;
		opt.useLFN		= FALSE;
		nohandlers		= TRUE;
	}

	if (!nohandlers) InstallExceptionHandlers();
xit:
	ppdefaults();
}

long ppgetvalue(char *s) {
	if (!*s)
		if (!ppargc--)
			ppinstructions();
		else
			return strtol(*ppargv++,NULL,0);
	return strtol(s,NULL,0);
}

void ppinstructions(void) {
	printf(	"%s\n\n"	/* version string */
		"Arguments:\n"
		"\t-w\tChoose picture width\n"
		"\t-h\tChoose picture height\n"
		"\t-de\tDebugging: Disable alternate exception handlers\n"
		"\t-dl\tDebugging: Do not use Windows 95 Long File Names\n"
		"\t-dL\tDebugging: Enable logging (must be first argument)\n"
		"\t-dv\tDebugging: Disable all VESA support\n"
		"\t-dv2\tDebugging: Disable VESA 2.0 support\n"
		"\t-dv3\tDebugging: Disable VESA 2.0 32-bit protected mode interface\n"
		"\t-dvl\tDebugging: Disable VESA Linear Frame Buffer support\n"
		"\t-dvm\tDebugging: Disable VESA mode search\n"
		"\t-dvp\tDebugging: Disable fast palette manipulation\n"
		"\t-dw\tDebugging: Start VP386 in wimp mode\n"
		"\t-mb\tMouse: Assume mouse driver allows signed coordinates\n"
		"\t-mp\tMouse: Turn off panning support\n",
		version_str);
	exit(10);
}

void ppdefaults(void) {
	if (!picbm->width) picbm->width=320;
	else picbm->width=(picbm->width+3)&0xFFFFFFFC;
	if (!picbm->height) picbm->height=200;
	else picbm->height=(picbm->height+3)&0xFFFFFFFC;
}



////////////////////////////////////////////////////////////////////////

//
//	boolean LoadResourceFile(char *rsrcname);
//
//	Loads VP386's Resource file into memory.
//
//	Returns TRUE on success, FALSE on failure.
//

static int sorter_KeyAliasNode(const struct KeyAliasNode *k1,const struct KeyAliasNode *k2) {
	return (int)(k1->key>k2->key)-(int)(k1->key<k2->key);
}

static int sorter_ResourceStrings(const ulong **k1,const ulong **k2) {
	return (int)(**k1>**k2) - (int)(**k1<**k2);
}




boolean LoadResourceFile(char *rsrcname) {
	FILE *f;
	char buf[19],*s,**sp;
	int c,i;
	ulong tl;

	if (!(f=VPfopen(rsrcname,"rb"))) return FALSE;

	for(;;) {
		if ((c=getc(f))==EOF) goto rerr;

		if (c) switch(c) {
		case 'k':	// keyboard shortcuts
			if (1!=fread(&key_shortcut_count,4,1,f)) goto rerr;
			if (!(key_shortcuts = malloc(sizeof(struct KeyAliasNode)*key_shortcut_count))) goto rerr;

			for(i=0;i<key_shortcut_count;i++) {
				memset(buf+7,0,12);
				if (1!=fread(buf,7,1,f)) goto rerr;
				if (buf[2]) if (1!=fread(buf+7,4*buf[2],1,f)) goto rerr;

				key_shortcuts[i].key	= *(ushort *)(buf+ 0);
				key_shortcuts[i].command= *(ulong  *)(buf+ 3);
				key_shortcuts[i].arg1	= *(ulong  *)(buf+ 7);
				key_shortcuts[i].arg2	= *(ulong  *)(buf+11);
				key_shortcuts[i].arg3	= *(ulong  *)(buf+15);
			} while(--i);

			qsort(key_shortcuts,key_shortcut_count,sizeof(struct KeyAliasNode),sorter_KeyAliasNode);

			break;
		case 's':
			if (1!=fread(&startup_strings.count,4,1,f)) goto rerr;
			if (1!=fread(&tl,4,1,f)) goto rerr;
			if (!(startup_strings.buf = malloc(tl))) goto rerr;
			if (1!=fread(startup_strings.buf,tl,1,f)) goto rerr;
			if (!(startup_strings.table = malloc(sizeof(char *)*startup_strings.count))) goto rerr;

			s = startup_strings.buf;
			sp = startup_strings.table;

			for(i=0;i<startup_strings.count;i++) {
				*sp++ = s;
				s+=4;
				while(*s++);
			}

			qsort(startup_strings.table,startup_strings.count,sizeof(char *),sorter_ResourceStrings);

			break;
		default:
			return FALSE;
		} else break;
	}

	return TRUE;
rerr:
	fclose(f);
	return FALSE;
}

//
//	char *FindResourceString(struct ResourceStrings *rscs,ulong num);
//
//	Finds a particular string in a String Resource.
//
//	Returns the string if found, or NULL if not found.
//

static int searcher_ResourceStrings(const ulong *k1,const ulong **k2) {
	return (int)(*k1>**k2) - (int)(*k1<**k2);
}

char *FindResourceString(struct ResourceStrings *rscs,ulong num) {
	void *key;

	if (key = bsearch(&num,rscs->table,rscs->count,sizeof(char *),searcher_ResourceStrings))
		return *(char **)key+4;
	else
		return NULL;
}

