#include "request.h"
#include "listbox.h"
#include <stdarg.h>

void ClipRequesterPos(struct Requester *req,long *x,long *y) {
	if (req->width >vidbm.width ) req->width =vidbm.width;
	if (req->height>vidbm.height) req->height=vidbm.height;
	if (*x+req->width >vidbm.x2) *x=vidbm.x2-req->width;
	if (*y+req->height>vidbm.y2) *y=vidbm.y2-req->height;
	if (*x<0) *x=0;
	if (*y<0) *y=0;
}

boolean OpenRequester(struct Requester *req,long x,long y) {
	int	barheight;

	++state.requester_count;

	if (req->flags & REQ_POPCEN) {
		mouse_coord mx,my,mb;

		mouse_read2(&mx,&my,&mb);
		x=mx-(req->width>>1);
		y=my-(req->height>>1);
	};

	if (req->flags & REQ_COORDBIAS)
		req->height += (req->flags & REQ_SYSTITLE) ? sysfontheight + 4*vdrv_cur->y_aspect - 1 : 11;

// suspend colorcycling

	if (!(req->flags & REQ_NOINHIBITCC))
		CCycle_Suspend();

// attempt to save the palette & force the colors
	
	if (req->flags & REQ_FORCECOL) {
		if (req->savepal=malloc(48+sizeof(struct StdColor))) {
			vid_saveRGB6(0,req->savepal,16);
			memcpy(req->savepal+48,&col,sizeof col);
			col = col_std;
			vid_loadRGB6(0,stdcol,16);
			req->flags |= REQ_COLCHANGED;
		}
	}

/* save the background underneath the requester */

	ClipRequesterPos(req,&x,&y);

	if (req->flags & REQ_NOSAVEBACK)
		req->saveback=NULL;
	else {
		mouse_hide();
		req->saveback=vd->blocksave(&vidbm,req->curx=x,req->cury=y,req->width,req->height);
		mouse_show();
	};

/* re-initialize the fields to point to this requester's bitmap */

	if (req->fields) {
		struct Field *f=req->fields;

		do {
			f->destbm=&req->reqbm;
		} while((++f)->x>=0);
	};

/* reset the requester by re-initializing the bitmap structure */
// if the flag REQ_COORDBIAS has been set, re-engineer so that the title
// is <(0,0)

	if (req->flags & REQ_COORDBIAS) {
		barheight = sysfontheight + 4*vdrv_cur->y_aspect - 1;
		vid_initBitMap(&req->reqbm,vidbm.vm+vidbm.bpr*(y+barheight)+x,0,-barheight,req->width,req->height,vidbm.bpr,vidbm.pal);
	} else {
		barheight = 11;
		vid_initBitMap(&req->reqbm,vidbm.vm+vidbm.bpr*y+x,0,0,req->width,req->height,vidbm.bpr,vidbm.pal);
	}

/* the VGAPaint 386 video library does not do full rectangle clipping --
** thus winbm should NEVER be used while a requester is being used.
** Therefore... we can do some *ahem* creative engineering...  This
** means that ClipOn() and ClipOff() cannot be called within a requester
** either without special treatment.
*/

	req->old_winbm=winbm;
	winbm=&req->reqbm;
	req->reqbm.syncrect=TRUE;
	winbm->syncx1=winbm->syncy1=0x7FFF;
	winbm->syncx2=winbm->syncy2=-0x8000;
	winbm->xbias=x;
	winbm->ybias=(req->flags & REQ_COORDBIAS) ? y+barheight : y;

	req->old_zonebm=zonebm;
	zonebm=&req->reqbm;

	mouse_hide();

	if (req->fields) {
		struct Field *f=req->fields;

		do
			f->destbm=&req->reqbm;
		while((++f)->x>=0);
	};

	if (req->sliders) {
		struct Slider *sld=req->sliders;

		do {
			sld->destbm=&req->reqbm;
			SliderInit(sld);
			sld=sld->next;
		} while(sld);
	};

	if (req->lboxes) {
		struct Listbox *lb=req->lboxes;

		do {
			lb->destbm=&req->reqbm;
			if (!lb->textheight) lb->textheight=sysfontheight;
			++lb;
		} while(lb->x != LB_END);
	};

	// render the requester

	RenderRequester(req);

	mouse_show();

	return TRUE;
}

void CloseRequester(struct Requester *req) {
	mouse_hide();

	winbm=req->old_winbm;
	
	if (req->sliders) {
		struct Slider *sld=req->sliders;

		do {
			SliderClose(sld);
			sld=sld->next;
		} while(sld);
	};

	if (!(req->flags & REQ_NOSAVEBACK))
		vd->blockrestore(&vidbm,req->saveback);

	vd->refresh();

	if (req->flags & REQ_COORDBIAS)
		req->height -= sysfontheight + 4*vdrv_cur->y_aspect - 1;

	mouse_show();

// if we changed the colors, change them back

	if (req->flags & REQ_COLCHANGED) {
		req->flags &= ~REQ_COLCHANGED;

		vid_loadRGB6(0,req->savepal,16);
		memcpy(&col,req->savepal+48,sizeof col);
		free(req->savepal);
	}

// resume colorcycling

	if (!(req->flags & REQ_NOINHIBITCC))
		CCycle_Resume();

	zonebm=req->old_zonebm;
	--state.requester_count;
}

ulong	CheckRequester(struct Requester *req,mouse_coord x,mouse_coord y,mouse_coord b) {
	ulong msg;

	/* is the point even in the requester!?!? */

	x-=req->curx;
	y-=req->cury;

	if (x<0 || y<0 || x>=req->width || y>=req->height) return REQM_NOTHING;

	/* any mouse button? */

	if (!(b & 7)) return REQM_NOTHING;

	/* first, check to see if the requester is in the drag bar */
	// only respond if the LMB was pressed

	if (b & 1) {
		if (req->title) {
			int barheight = (req->flags & REQ_SYSTITLE) ? sysfontheight + 4*YPOINT : 11;

			if (y<((req->flags & REQ_SYSTITLE) ? sysfontheight + 4*YPOINT : 11)) {
				if (req->flags & REQ_NODRAG)
					mouse_wait_unkey();
				else
					DragRequester(req);
				return REQM_DRAGGED;
			}
		}
	}

	// bias the y coordinate correctly

	if (req->flags & REQ_COORDBIAS)
		y -= (req->flags & REQ_SYSTITLE ? sysfontheight+4*YPOINT-1 : 11);

	/* second, check all zones */

	if (-1!=(msg=ZoneRangeAct(req->zones,x,y)))
		return REQM_ZONE+msg;

	/* third, check all fields */

	if (req->fields) {
		struct Field *f=req->fields;

		msg=REQM_FIELD;

		do {
			if (rectInside((struct Rectangle *)f,x,y)) {
				ActivateField(f,x);
				return msg;
			};
			++msg;
		} while((++f)->x >=0);
	};

	/* fourth, check all sliders */

	if (req->sliders) {
		struct Slider *sld=req->sliders;
		ulong	ctick,ntick;

		ctick = *(ulong*)0x46CL;
		ntick = ctick+9;

		do {
			if (msg=SliderHandle(sld,x,y)) {
				if (sld->flags & REQSLD_ONEHIT) {
					if ((msg>=2 && msg<6) || (msg>=7 && msg<11))
						mouse_wait_unkey();
				} else {
					while(((slong)(ntick-ctick))>0 && (mouse_read2(&x,&y,&b),b)) {
						ctick = *(ulong *)0x46CL;
						kbhit();
					}
					x-=req->curx;
					y-=req->cury;
					if (req->flags & REQ_COORDBIAS)
						y -= (req->flags & REQ_SYSTITLE ? sysfontheight+4*YPOINT-1 : 11);
					if (b) {
						ntick = ctick+1;
						continue;
					}
				}
				return msg+REQM_SLIDER;
			};

			msg+=0x10000;
			sld=sld->next;
		} while(sld);
	};

	/* fifth, check all listboxes */

	if (req->lboxes) {
		struct Listbox *lb=req->lboxes;

		msg=REQM_LISTBOX;
		do {
			if (rectInside((struct Rectangle *)lb,x,y)) {
				SelectListbox(lb,lb->cpos+(y-lb->y)/lb->textheight);
				return msg;
			};
			msg++,lb++;
		} while(lb->x != LB_END);
	};

	return REQM_NOTHING;
}

void MoveRequester(struct Requester *req,long x,long y) {
	mouse_hide();
	if (req->saveback) {
		vd->bitbltswap(&vidbm,req->curx,req->cury,&req->saveback->bm,0,0,req->width,req->height);
		vd->bitbltswap(&vidbm,req->curx=x,req->cury=y,&req->saveback->bm,0,0,req->width,req->height);
		req->saveback->x=x;
		req->saveback->y=y;
	} else {
		vd->bitblt(&vidbm,req->curx,req->cury,&vidbm,x,y,req->width,req->height);
	};

	/* recompute the bitmap's base address */

	req->reqbm.xbias=x;
	if (req->flags & REQ_COORDBIAS) {
		int y2 = y+(req->flags & REQ_SYSTITLE ? sysfontheight + 4*YPOINT - 1 : 11);

		req->reqbm.vm=vidbm.vm+vidbm.bpr*y2+x;
		req->reqbm.ybias=y2;
	} else {
		req->reqbm.vm=vidbm.vm+vidbm.bpr*y+x;
		req->reqbm.ybias=y;
	}

	vd->refresh();

	mouse_show();
}

void DragRequester(struct Requester *req) {
	mouse_coord x,y,b;
	long offx,offy,ox,oy;

	mouse_read2(&x,&y,&b);
	offx=x-req->curx;
	offy=y-req->cury;
	x=req->curx;
	y=req->cury;

	do {
		if (!opt.reqImageDrag) {
			mouse_hide();
			vd->box(&vidbm,x,y,x+req->width-1,y+req->height-1,-1);
			vd->refresh();
			mouse_show();
		}
		ox=x; oy=y;
		do {
			mouse_read2(&x,&y,&b);
			x-=offx;
			y-=offy;
			if (x<0) x=0; else if (x>vidbm.width -req->width ) x=vidbm.width -req->width;
			if (y<0) y=0; else if (y>vidbm.height-req->height) y=vidbm.height-req->height;
		} while((b&1) && ox==x && oy==y);
		if (opt.reqImageDrag) {
			if (x!=req->curx || y!=req->cury)
				MoveRequester(req,x,y);
		} else {
			mouse_hide();
			vd->box(&vidbm,ox,oy,ox+req->width-1,oy+req->height-1,-1);
			mouse_show();
		}
	} while(b & 1);

	MoveRequester(req,x,y);
}

//	void	RenderRequester(struct Requester *req);
//
//	Completely re-render a requester from scratch.

void	RenderRequester(struct Requester *req) {
	int	barheight,bary;

	barheight = sysfontheight + 4*vdrv_cur->y_aspect - 1;
	bary	  = 2*vdrv_cur->y_aspect;

	vd->set(&req->reqbm,req->backfill);		// erase the background

	if (req->flags & REQ_BORDER) {			// re-render title
		if (req->flags & REQ_COORDBIAS)
			vid_draw3d(&req->reqbm,0,-barheight,req->width-1,req->height-1-barheight);
		else
			vid_draw3d(&req->reqbm,0,0,req->width-1,req->height-1);
	};

	Render3DList(&req->reqbm,req->pblist,TRUE);	// render positive boxes
	Render3DList(&req->reqbm,req->nblist,FALSE);	// render negative boxes
	RenderTextList(&req->reqbm,req->tlist);		// render text list

	// if a title was included we are obliged to render it

	if (req->title) {
		if (req->flags & REQ_COORDBIAS) {
			if (req->flags & REQ_SYSTITLE) {
				SysText(&req->reqbm,(req->width+1-sysfontwidth*(long)strlen(req->title))>>1,bary-barheight,req->title,
					ST0,req->backfill);
				vid_draw3d(&req->reqbm,0,-barheight,req->width-1,-1);
			} else {
				vd->text6(&req->reqbm,(req->width+1-6*strlen(req->title))>>1,2-11,req->title,ST0,req->backfill);
				vid_draw3d(&req->reqbm,0,-11,req->width-1,-1);
			}
		} else {
			if (req->flags & REQ_SYSTITLE) {
				SysText(&req->reqbm,(req->width+1-sysfontwidth*strlen(req->title))>>1,bary,req->title,
					ST0,req->backfill);
				vid_draw3d(&req->reqbm,0,0,req->width-1,barheight-1);
			} else {
				vd->text6(&req->reqbm,(req->width+1-6*strlen(req->title))>>1,2,req->title,ST0,req->backfill);
				vid_draw3d(&req->reqbm,0,0,req->width-1,11-1);
			}
		}
	};

	if (req->zones) {
		ZoneRangeInit(req->zones);
		ZoneRefresh(req->zones);
	}

	if (req->fields) {
		struct Field *f=req->fields;

		do
			DisplayField(f);
		while((++f)->x>=0);
	};

	if (req->sliders) {
		struct Slider *sld=req->sliders;

		do {
			SliderRedraw(sld);
			SliderDraw(sld);
			sld=sld->next;
		} while(sld);
	};

	if (req->lboxes) {
		struct Listbox *lb=req->lboxes;

		do
			DrawListbox(lb++);
		while(lb->x != LB_END);
	};

	vd->refresh();		// for good measure
}

//	void HideRequester(struct Requester *req);
//
//	Hide a requester so that it no longer shows on screen.  Changed
//	system values are restored so that underlying requesters/gadgets
//	can be accessed again.
//
//	This is only valid if the requester has a saveback pointer.  If
//	the background wasn't saved, this routine will abort.

void	HideRequester(struct Requester *req) {
	if (req->flags & REQ_HIDDEN) return;
	if (req->flags & REQ_NOSAVEBACK) return;

	// flip the bitmap back

	mouse_hide();
	vd->bitbltswap(&req->saveback->bm,0,0,&vidbm,req->curx,req->cury,req->saveback->xs,req->saveback->ys);
	vd->refresh();
	mouse_show();

	// restore winbm & zonebm
	winbm = req->old_winbm;
	zonebm = req->old_zonebm;

	req->flags |= REQ_HIDDEN;
}

//	void ShowRequester(struct Requester *req);
//
//	Shows a previously hidden requester.  System values restored by
//	HideRequester() are taken over again.

void	ShowRequester(struct Requester *req) {
	if (!(req->flags & REQ_HIDDEN)) return;

	// flip the bitmap back

	mouse_hide();
	vd->bitbltswap(&req->saveback->bm,0,0,&vidbm,req->curx,req->cury,req->saveback->xs,req->saveback->ys);
	vd->refresh();
	mouse_show();

	// regrab winbm & zonebm
	winbm = zonebm = &req->reqbm;

	req->flags &= ~REQ_HIDDEN;
}



/////////////////// STANDARD REQUESTER ROUTINES //////////////////////


long	reqsld_bias;

void	RequestSlider_Updater(struct Slider *sld) {
	mouse_hide();
	vid_printf6(sld->destbm,101-6,12,St0,St2," %7d ",sld->cposx-reqsld_bias);
	mouse_show();
}


static struct Slider reqsld_sliders[]={
	{
		NULL,
		{ NULL },
		NULL,		/* bitmap will be filled in */
		{ 24,22,208,8 },
		NULL,		/* use autoknob */
		0,0,		/* will be filled in */
		4,4,		/* knob sizes */
		0,0,		/* listbox.c */
		255,0,		/* clicks */
		16,16,		/* bump */
		0,0,0,0,	/* positions */
		SLD_ARROW1*4 + SLD_ARROW2*6 + SLD_OUTLINE3D + SLD_ATTACHX + SLD_ATTACHY,	// flags
		RequestSlider_Updater,	// updater
		{  10,20,12,12 },	// arrow1
		{ 234,20,12,12 },	// arrow2
		-1,0,1,0,		// arrow taps
		NULL			// no controlled listbox
	}
};

static struct Requester req_sld={
	0,0,
	256,36,
	REQ_BORDER + REQ_POPCEN + REQ_FORCECOL + REQ_SLDREPHIT,
	NULL,		/* no title */
	St2,
	NULL,
	NULL,
	NULL,
	&reqsld_sliders,
	NULL,
	NULL,			// negatives
	NULL
};

long	RequestSlider(char *title,long bias,ulong range,long lastval) {
	mouse_coord x,y,b;
	int		key;

	reqsld_bias = bias;
	reqsld_sliders[0].xclicks = range;

	req_sld.title=title;
	if (!OpenRequester(&req_sld,0,0)) return range/2-bias;

	SliderAbs(&reqsld_sliders[0],lastval+bias,0);

	for(;;) {
		mouse_wait_unkey();
		do
			kbhit();
		while(!mouse_buttons() && -1==(key = kbd_check()));
		if (key==0x1b || key==' ') break;
		mouse_read2(&x,&y,&b);
		if (b&2) break;
		CheckRequester(&req_sld,x,y,b);
	};
	
	mouse_wait_unkey();

	CloseRequester(&req_sld);
	return reqsld_sliders[0].cposx-bias;
}


struct SliderRequester *MakeSliderRequester(long width) {
	struct SliderRequester *sldr;

	if (!(sldr = malloc(sizeof(struct SliderRequester)))) return NULL;

	memcpy(&sldr->req,&req_sld,sizeof(struct Requester));
	memcpy(&sldr->sld,&reqsld_sliders[0],sizeof(struct Slider));

	sldr->req.flags		= REQ_BORDER+REQ_POPCEN+REQ_FORCECOL+REQ_SLDREPHIT+REQ_COORDBIAS+REQ_SYSTITLE;
	sldr->req.sliders	= &sldr->sld;
	sldr->req.width		= width*XPOINT;
	sldr->req.height	= sysfontheight+12*YPOINT+4;
	sldr->sld.boundbox.x	= 16*XPOINT;
	sldr->sld.boundbox.y	= sysfontheight+4*YPOINT;
	sldr->sld.boundbox.width= XPOINT*(width-32);
	sldr->sld.boundbox.height=8*YPOINT;
	sldr->sld.arrow1.width = sldr->sld.arrow2.width = 12*XPOINT;
	sldr->sld.arrow1.height = sldr->sld.arrow2.height = 12*YPOINT;

	return sldr;
}

void	DisposeSliderRequester(struct SliderRequester *sldr) {
	free(sldr);
}


///////// boolean requester


static struct Zone reqbool_zones[]={
	{ 2,22,0,13,ZN_3D+ZN_TEXTRENDER+ZN_SYSTEXT,NULL,NULL,NULL },
	{ -1,-1,-1,-1 },
	{ 0,22,0,13,ZN_3D+ZN_TEXTRENDER+ZN_SYSTEXT,NULL,NULL,NULL },
	{ -1,-1,-1,-1 },
	{ 0,22,0,13,ZN_3D+ZN_TEXTRENDER+ZN_SYSTEXT,NULL,NULL,NULL },
	{ -1,-1,-1,-1 }
};

static struct ZoneRange reqbool_zoneranges[]={
	{ &reqbool_zoneranges[1],&reqbool_zones[0],NULL,-1,ZR_TAP,NULL,-1,0,NULL,0 },
	{ &reqbool_zoneranges[2],&reqbool_zones[2],NULL,-1,ZR_TAP,NULL,-1,0,NULL,0 },
	{ NULL			,&reqbool_zones[4],NULL,-1,ZR_TAP,NULL,-1,0,NULL,0 }
};

//static struct Text reqbool_texts[]={
//	{ NULL,0,12,St0,St2 },
//	{ NULL }
//};

static struct Requester reqbool_req={
	0,0,320,38,
	REQ_BORDER+REQ_POPCEN+REQ_FORCECOL+REQ_SYSTITLE,
	NULL,
	St2,
	NULL,NULL,reqbool_zoneranges,NULL,NULL,NULL,
//	reqbool_texts,NULL
	NULL,NULL
};

short	RequestBoolean(char *title,char *yes,char *maybe,char *no,char *format, ...) {
	va_list val;
	mouse_coord mx,my,mb;
	short i;
	boolean quit=FALSE;
	char buf[128];

	reqbool_req.title=title;
	va_start(val,format);
	vsprintf(buf,format,val);
	va_end(val);

	i=(yes ? strlen(yes):0)+(no?strlen(no):0)+(maybe?strlen(maybe):0);

	reqbool_req.width=max(max(strlen(title)*sysfontwidth+4*XPOINT,i*sysfontwidth+16*XPOINT),strlen(buf)*sysfontwidth+4*XPOINT);

	reqbool_zones[0].y  = reqbool_zones[2].y  = reqbool_zones[4].y  = sysfontheight*2 + 6*YPOINT;
	reqbool_zones[0].ys = reqbool_zones[2].ys = reqbool_zones[4].ys = sysfontheight + 4*YPOINT;
	reqbool_req.height = sysfontheight*3 + 12*YPOINT;

	reqbool_zones[0].xs=reqbool_zones[2].xs=-1;
	reqbool_zones[4].text=no;
	reqbool_zones[4].xs=strlen(no)*sysfontwidth+4*XPOINT;
	reqbool_zones[4].x=reqbool_req.width-reqbool_zones[4].xs-2*XPOINT;
	if (yes) {
		reqbool_zones[0].text=yes;
		reqbool_zones[0].xs=strlen(yes)*sysfontwidth+4*XPOINT;
	};
	if (maybe) {
		reqbool_zones[2].text=maybe;
		reqbool_zones[2].xs=strlen(maybe)*sysfontwidth+4*XPOINT;
		reqbool_zones[2].x=(reqbool_zones[0].x+reqbool_zones[0].xs+reqbool_zones[4].x-reqbool_zones[2].xs)>>1;
	};

	if (!OpenRequester(&reqbool_req,0,0)) return 0;	/* cancel */

	mouse_hide();
	SysText(&reqbool_req.reqbm,(reqbool_req.width-strlen(buf)*sysfontwidth)>>1,
		sysfontheight + 5*vdrv_cur->y_aspect,buf,St0,St2);
	vd->refresh();
	mouse_show();

	do {
		if (kbhit()) {
			switch(getch()) {
			case '1':		// 1
			case 'y': case 'Y':	// Y
				i=2; quit=TRUE; continue;	// almost a dbl-break...
			case '2':		// 2
			case 'm': case 'M':	// M
				i=1; quit=TRUE; continue;
			case 0x1B:		// ESC
			case '3':		// 3
			case 'n': case 'N':	// N
				i=0; quit=TRUE; continue;
			};
		};
		mouse_read2(&mx,&my,&mb);
		switch(CheckRequester(&reqbool_req,mx,my,mb)) {
		case RZoneMsg(0,0):	i=2; quit=TRUE; break;
		case RZoneMsg(1,0):	i=1; quit=TRUE; break;
		case RZoneMsg(2,0):	i=0; quit=TRUE; break;
		};
	} while(!quit);

	CloseRequester(&reqbool_req);

	return i;
}

#pragma off (check_stack);

////// status bar "requester"
//
//	these are designed so that you may call CreateStatusBar() without
//	checking for a NULL result, so just do an sb=CreateStatusBar() and
//	then go ahead and call ModifyStatusBar() and DestroyStatusBar() without
//	any checking...

static struct Rectangle CSB_Negs[]={
	{ 4,13,314,6 },
	{RECT_END}
};

static struct Requester CSB_Req={
	0,0,320,22,REQ_BORDER+REQ_NODRAG+REQ_POPCEN,
	NULL,
	St2,
	NULL,NULL,NULL,NULL,		// lboxes fields zones sliders
	NULL,CSB_Negs,NULL,NULL		// pos neg tlist oldwinbm
};

struct StatusBar *	CreateStatusBar(long maxval,char *format, ...) {
	struct StatusBar *sb;
	va_list val;

	if (sb=malloc(sizeof(struct StatusBar))) {
		memset(sb,0,sizeof(struct StatusBar));
		sb->maximum=maxval;
		if (sb->req=malloc(sizeof(struct Requester))) {
			*sb->req=CSB_Req;
			if (sb->req->title=malloc(1024)) {
				va_start(val,format);
				vsprintf(sb->req->title,format,val);
				va_end(val);
				realloc(sb->req->title,strlen(sb->req->title)+1);

				if (OpenRequester(sb->req,0,0)) return sb;

				// failed somewhere

				free(sb->req->title);
			};
			free(sb->req);
		};
		free(sb);
	};
	return NULL;
}

void	ModifyStatusBar(struct StatusBar *sb,long statuspt) {
	coord newpixels;

	if (!sb) return;

	if (statuspt<=sb->atpoint) return;

	newpixels=(statuspt*310+(sb->maximum>>1))/sb->maximum;
	if (newpixels==sb->curpixels) return;

	mouse_hide();
	vd->bar(&sb->req->reqbm,5+sb->curpixels,14,newpixels+1-sb->curpixels,4,St1);
	vd->refresh();
	mouse_show();

	sb->curpixels=newpixels;
}

void	DestroyStatusBar(struct StatusBar *sb) {
	if (!sb) return;

	CloseRequester(sb->req);
	free(sb->req->title);
	free(sb->req);
	free(sb);
}

////////////////////////// INFORMATION LISTBOX REQUESTER /////////////////////////

// Whoopee, my first dynamic requester!  This ought to be fun...

struct ListRequester *MakeListRequester(long xsiz,long ysiz,char *name) {
	struct ListRequester *lr;

	if (!(lr	=NEW(struct ListRequester)))	return NULL;
	memset(lr,0,sizeof(struct ListRequester));
	if (!(lr->req	=NEW(struct Requester)))	goto abort;
	if (!(lr->tbox	=NEW(struct Textbox)))		goto abort;
	if (!(lr->lbox	=ANEW(struct Listbox,2)))	goto abort;
	if (!(lr->zones	=ANEW(struct Zone,3)))		goto abort;
	if (!(lr->zr	=NEW(struct ZoneRange)))	goto abort;
	if (!(lr->sld	=ANEW(struct Slider,2)))	goto abort;
	if (!(lr->neg	=ANEW(struct Rectangle,2)))	goto abort;

	// clear structures

	memset(lr->req,0,sizeof(struct Requester));
	memset(lr->lbox,0,sizeof(struct Listbox [2]));
	memset(lr->zones,0,sizeof(struct Zone [3]));
	memset(lr->zr,0,sizeof(struct ZoneRange));
	memset(lr->sld,0,sizeof(struct Slider));

	// link in all these structures

	lr->tbox->lb		= lr->lbox;
	lr->req->flags		= REQ_BORDER+REQ_POPCEN+REQ_FORCECOL+REQ_COORDBIAS+REQ_SYSTITLE;
	lr->req->title		= name;
	lr->req->backfill	= St2;
	lr->req->lboxes		= lr->lbox;
	lr->req->zones		= lr->zr;
	lr->req->sliders	= lr->sld;
	lr->req->nblist		= lr->neg;
	lr->lbox[0].x		= 6*XPOINT;
	lr->lbox[0].y		= 5*YPOINT;
	lr->lbox[1].x		= LB_END;
	lr->zones[0].x		= 4*XPOINT;
	lr->zones[0].xs		= lr->zones[1].xs	= 4*XPOINT+8*sysfontwidth;
	lr->zones[0].ys		= lr->zones[1].ys	= 4*YPOINT+sysfontheight;
	lr->zones[0].text	= "Ok";
	lr->zones[1].text	= "Cancel";
	lr->zones[0].flags	= lr->zones[1].flags = ZN_3D+ZN_TEXTRENDER+ZN_SYSTEXT;
	memset(&lr->zones[2],0xFF,(char*)&lr->zones[2].flags - (char*)&lr->zones[2]);	// -1,-1,-1,-1
	lr->zr->zonelist	= lr->zones;
	lr->zr->highindex	= -1;
	lr->zr->flags		= ZR_TAP;
	lr->sld->boundbox.y	= 17*YPOINT;
	lr->sld->boundbox.width	= 8*XPOINT;
	lr->sld->min_width	= lr->sld->min_height	= 4;
	lr->sld->xbump		= lr->sld->ybump	= 4;
	lr->sld->flags		= REQSLD_ONEHIT + SLD_ARROW1HIT + SLD_ARROW2HIT + SLD_ARROW1*2 + SLD_ARROW2*8
				+ SLD_OUTLINE3D + SLD_LISTBOX + SLD_ATTACHY;
//	lr->sld->arrow1.y	= 14*YPOINT;
	lr->sld->arrow1.width	= lr->sld->arrow2.width = 12*XPOINT;
	lr->sld->arrow1.height	= lr->sld->arrow2.height= 12*YPOINT;
	lr->sld->arrow1tapy	= -1;
	lr->sld->arrow2tapy	= 1;
	lr->sld->lboxctrl	= &lr->lbox[0];
	lr->neg[0].x		= 4*XPOINT;
	lr->neg[0].y		= 3*YPOINT;
	lr->neg[1].x		= RECT_END;

	ResizeListRequester(lr,xsiz,ysiz);

	// hurrah!

	return lr;

abort:
	DisposeListRequester(lr);
	return NULL;
}

void	ResizeListRequester(struct ListRequester *lr,coord xsiz,coord ysiz) {
	lr->width		= xsiz;
	lr->height		= ysiz;
	lr->req->width		= 24*XPOINT + sysfontwidth*xsiz;
	lr->req->height		= 16*YPOINT + sysfontheight*(ysiz+1);
	lr->lbox[0].width	= sysfontwidth*xsiz;
	lr->lbox[0].height	= sysfontheight*ysiz;
	lr->lbox[0].cwidth	= xsiz;
	lr->lbox[0].cheight	= ysiz;
	lr->lbox[0].textheight	= sysfontheight;
	lr->zones[0].y		= lr->zones[1].y	= 8*YPOINT + sysfontheight*ysiz;
	lr->zones[1].x		= sysfontwidth*xsiz + 16*XPOINT - 8*sysfontwidth;
	lr->sld->boundbox.x	= 10*XPOINT + sysfontwidth*xsiz;
	lr->sld->boundbox.height= sysfontheight*ysiz - 24*YPOINT;
//	lr->sld->arrow1.x	= lr->sld->arrow2.x	= 8*XPOINT + sysfontwidth*xsiz;
//	lr->sld->arrow2.y	= 6*YPOINT + sysfontheight*ysiz;
	lr->neg[0].width	= 4*XPOINT + sysfontwidth*xsiz;
	lr->neg[0].height	= 4*YPOINT + sysfontheight*ysiz;
	InitTextbox(lr->tbox,lr->lbox);
}

boolean	OpenListRequesterAutosize(struct ListRequester *lr) {
	int	ys = (vidbm.height - 19*YPOINT - 1)/sysfontheight-2;

	while(ys>0) {
		ResizeListRequester(lr,lr->width,ys);
		if (OpenRequester(lr->req,0,0)) return TRUE;
		--ys;
	}
	return FALSE;
}

long	ActivateListRequesterAutosize(struct ListRequester *lr) {
	if (OpenListRequesterAutosize(lr))
		return ActivateListRequester1(lr);
	return lr->lbox->sel;
}

long	ActivateListRequester(struct ListRequester *lr) {
	if (!OpenRequester(lr->req,0,0)) return lr->lbox->sel;

	return ActivateListRequester1(lr);
}

long	ActivateListRequester1(struct ListRequester *lr) {
	mouse_coord	mx,my,mb;
	ulong		msg;
	boolean		cont=TRUE;
	boolean		okay = FALSE;

	do {
		mouse_read2(&mx,&my,&mb);
		if (mb & 1) {
			msg=CheckRequester(lr->req,mx,my,mb);
			mouse_wait_unkey();
			switch(msg) {
			case RZoneMsg(0,0):	/* Ok */
				okay = TRUE;
			case RZoneMsg(0,1):	/* Cancel */
				cont=FALSE;
				break;
			};
		};
	} while(cont);

	CloseRequester(lr->req);

	return okay ? lr->lbox->sel : -1;
}

void	DisposeListRequester(struct ListRequester *lr) {
	free(lr->req);
	free(lr->tbox);
	free(lr->lbox);
	free(lr->zones);
	free(lr->zr);
	free(lr->sld);
	free(lr->neg);
	free(lr);
}

////////////////////// REQUESTER RENDERING ROUTINES /////////////////////////


void Render3DList(BMP bm,struct Rectangle *rl,boolean pos) {
	void (*d3d)(BMP,coord,coord,coord,coord);

	if (!rl) return;

	d3d=pos ? &vid_draw3d : &vid_inv3d;

	while(rl->x != RECT_END) {
		d3d(bm,rl->x,rl->y,rl->x+rl->width-1,rl->y+rl->height-1);
		++rl;
	};
	vd->refresh();	/* if you've got time to do this you've got time for a refresh */
}

void RenderTextList(BMP bm,struct Text *tl) {
	char *s;

	if (!tl) return;

	while(s=tl->s) {
		if (tl->back!=0xFFFF)
			vd->text6(bm,tl->x,tl->y,s,tl->fore,tl->back);
		else
			vd->text6t(bm,tl->x,tl->y,s,tl->fore);
		++tl;
	};
	vd->refresh();	/* look above... */
}

