#define	VGAPAINT_SPRITE_C

#include <mem.h>
#include <string.h>

#include "sprite.h"
#include "video.h"
#include "vgapaint.h"

struct Sprite dummy_sprite={NULL,NULL};

/* sprite_init: initialize a sprite for use */

void sprite_init(struct Sprite *spr,uchar *image,uchar *savemem,struct BitMap *dstbm,long imagebpr) {
	vid_initBitMap(&spr->savebm,savemem,0,0,spr->width,spr->height,imagebpr,dstbm->pal);
	vid_initBitMap(&spr->imgbm,image,0,0,spr->width,spr->height,imagebpr,dstbm->pal);
	sprite_reinit(spr);
	spr->flags &= ~(SPR_DISPLAYED | SPR_GLOBAL | SPR_HIDDEN);
	spr->dispx=spr->dispy=-0x8000;
	spr->dstbm=dstbm;
	return;
}

/* sprite_reinit: reinitialize a sprite once the image has changed */

void sprite_reinit(struct Sprite *spr) {
//	if (spr->savebm.vm != spr->imgbm.vm)
//		memcpy(spr->savebm.vm,spr->imgbm.vm,spr->width * spr->height);
}

/* sprite_add: add a sprite to the global sprite list */

void sprite_add(struct Sprite *spr) {
	if (!(spr->flags & SPR_GLOBAL)) {
		spr->prev=NULL;
		spr->next=dummy_sprite.next;
		dummy_sprite.next=spr;
		if (!dummy_sprite.prev) dummy_sprite.prev=spr;
		spr->flags |= SPR_GLOBAL;
	};
}

/* sprite_rem: remove a sprite from the global sprite list */

void sprite_rem(struct Sprite *spr) {
	if (spr->flags & SPR_GLOBAL) {
		if (spr->next) spr->prev->next=spr->next;
		if (spr->prev) spr->next->prev=spr->prev;
		spr->flags &= ~SPR_GLOBAL;
	};
}

/* sprite_draw: draw a sprite using bitswap method */

void sprite_draw(struct Sprite *spr) {
	if (spr->flags & SPR_DISPLAYED) return;
	if (!(spr->flags & SPR_HIDDEN)) {
		spr->dispx=spr->x;
		spr->dispy=spr->y;

		vd->bitblt(spr->dstbm,spr->x,spr->y,&spr->savebm,0,0,spr->width,spr->height);

		switch (spr->flags & SPR_DRAWMODE) {
		case SPR_NOPATTERN:
			vd->bitbltcolor(&spr->imgbm,0,0,spr->dstbm,spr->x,spr->y,spr->width,spr->height,spr->trans_colr,spr->colr);
			break;
		case SPR_MATTE:
			vd->bitblttrans(&spr->imgbm,0,0,spr->dstbm,spr->x,spr->y,spr->width,spr->height,spr->trans_colr);
			break;
		case SPR_REPLACE:
			vd->bitblt(&spr->imgbm,0,0,spr->dstbm,spr->x,spr->y,spr->width,spr->height);
			break;
		case SPR_NOIMAGE:
			vd->bar(spr->dstbm,spr->x,spr->y,spr->width,spr->height,spr->colr);
			break;
		};
		spr->flags |= SPR_DISPLAYED;
	};
}

/* sprite_erase: erase a sprite using bitswap method */

void sprite_erase(struct Sprite *spr) {
	if (!(spr->flags & SPR_DISPLAYED)) return;
	vd->bitblt(&spr->savebm,0,0,spr->dstbm,spr->dispx,spr->dispy,spr->width,spr->height);
	spr->flags &= ~SPR_DISPLAYED;
}

/* sprite_drawall: draw all sprites on screen, forward order */

void sprite_drawall(void) {
	struct Sprite *spr;

	spr=dummy_sprite.prev;

	while(spr) {
		sprite_draw(spr);
		spr=spr->prev;
	};
}

/* sprite_eraseall: erase all sprites on screen, backward order */

void sprite_eraseall(void) {
	struct Sprite *spr;

	spr=dummy_sprite.next;

	while(spr) {
		sprite_erase(spr);
		spr=spr->next;
	};
}

/* sprite_hide: set hidden bit of a sprite and erase it if necessary
		automatically erases entire list and redraws if necessary */

void sprite_hide(struct Sprite *spr) {
	if (spr->flags & SPR_GLOBAL)
		sprite_eraseall();
	spr->flags |= SPR_HIDDEN;
	if (spr->flags & SPR_GLOBAL)
		sprite_drawall();
	else
		sprite_erase(spr);
}

/* sprite_hide2: hide a sprite quickly & handle mouse ptr */

void sprite_hide2(struct Sprite *spr) {
	spr->flags |= SPR_HIDDEN;
	if (spr->flags & SPR_DISPLAYED) {
		mouse_hide();
		sprite_erase(spr);
		mouse_show();
	};
}

/* sprite_show: reset hidden bit of a sprite and redraw it if necessary
		automatically erases entire list and redraws if necessary */

void sprite_show(struct Sprite *spr) {
	if (spr->flags & SPR_GLOBAL)
		sprite_eraseall();
	spr->flags &= ~SPR_HIDDEN;
	if (spr->flags & SPR_GLOBAL)
		sprite_drawall();
	else
		sprite_draw(spr);
	vd->refresh();
}

/* sprite_show2: show sprite quickly & handle mouse ptr */

void sprite_show2(struct Sprite *spr) {
	spr->flags &= ~SPR_HIDDEN;
	if (!(spr->flags & SPR_DISPLAYED)) {
		mouse_hide();
		sprite_draw(spr);
		mouse_show();
	};
}

/* sprite_quikmove: quickly move a sprite using erase-move-draw */

void sprite_quikmove(struct Sprite *spr,short x,short y) {
	if (x==spr->dispx && y==spr->dispy) return;
	sprite_erase(spr);
	spr->x=x;
	spr->y=y;
	sprite_draw(spr);
}

/* sprite_quikmove2: quickly move a sprite, but manage mouse pointer */

void sprite_quikmove2(struct Sprite *spr,short x,short y) {
	if (x==spr->dispx && y==spr->dispy) return;
	spr->x=x;
	spr->y=y;
	mouse_hide();
	sprite_erase(spr);
	sprite_draw(spr);
	mouse_show();
	vd->refresh();
}

/* sprite_paste: draw a sprite on screen w/o regard for erasure */

void sprite_paste(struct BitMap *bm,struct Sprite *spr,short x,short y) {
	switch(spr->flags & SPR_DRAWMODE) {
	case SPR_NOPATTERN:
		vd->bitbltcolor(&spr->imgbm,0,0,bm,x,y,spr->width,spr->height,spr->trans_colr,spr->colr);
		break;
	case SPR_MATTE:
		vd->bitblttrans(&spr->imgbm,0,0,bm,x,y,spr->width,spr->height,spr->trans_colr);
		break;
	case SPR_REPLACE:
		vid13_bitblt(&spr->imgbm,0,0,bm,x,y,spr->width,spr->height);
		break;
	case SPR_NOIMAGE:
		vd->bar(bm,x,y,spr->width,spr->height,spr->colr);
		break;
	};
	vd->refresh();
}

/* sprite_pointcoll: test for collision between a point and a sprite */

boolean sprite_pointcoll(struct Sprite *spr,short x,short y) {
	if ((x-=spr->x)>=0)
		if ((y-=spr->y)>=0)
			if (x<spr->width)
				if (y<spr->height)
					return !!spr->imgbm.vm[y*spr->width+x];
	return FALSE;
}

void	sprite_3dtranslate(uchar *dst,uchar *src,ulong size) {
	while(size--) *dst++=(uchar)(((color *)&col)[*src++]);
}
