#include <a/types.h>
#include <a/mem.h>
#include "loadlib.h"
#include "toolicon.h"
#include "video.h"
#include "extdata.h"

boolean	ovid_init(enum VD_InitCode vdic,long x,long y);
void	ovid_rbitblt(coord x1,coord y1,coord xs,coord ys,uchar *mem,long dmod,boolean no_offset);
void	ovid_rhline(coord x,coord y,uchar *mem,coord xs);
void	ovid_rhlineread(coord x,coord y,uchar *mem,coord xs);
void	ovid_drawptr(struct MousePointerInfo *mpi,coord x,coord y);
void	ovid_eraseptr(struct MousePointerInfo *mpi);

struct VDevSpecs ovds={
	320,200,320,200,320,200,320,200
};

struct VDev ovd={
	NULL,				// set
	NULL,				// plot
	NULL,				// point
	NULL,				// bar
	NULL,				// bitblt
	NULL,				// bitbltcolor
	NULL,				// bitblttrans
	NULL,				// bitbltswap
	NULL,				// bitbltswaptrans
	NULL,				// bitbltswapmask,
	NULL,				// bltpattern
	NULL,				// bltpattern3d
	NULL,				// text6
	NULL,				// text6trans
	NULL,				// text6lim,
	NULL,				// hline
	NULL,				// vline
	NULL,				// line
	NULL,				// box
	NULL,				// circle
	NULL,				// ellipse
	NULL,				// blocksave
	NULL,				// blockrestore
	NULL,				// hblit
	NULL,				// vblit
	NULL,				// tile
	NULL,				// rassize
	NULL,				// bpr
	&ovid_init,
	NULL,				// refresh
	&ovid_rhline,
	NULL,				// quickfill
	&ovid_rbitblt,
	&ovid_drawptr,
	&ovid_eraseptr,
	&ovid_rhlineread
};

struct VideoDriver ourVideoDriver={
	{NULL},"320x200 dblhigh paged",&ovd,NULL,VDF_FONT6x8+VDF_RATIO,256,VDEV_SIZE,VDRV_SIZE,1,1,0x238e2,0x07333
};

static long scroll_y=0;	// current y position

long main(struct Extension *ext,enum ExtCommand comm,long xval,long yval) {
	switch(comm) {
	case	EC_STARTUP:
		return (long)&main;

	case	EC_STARTUP1:
		// set our name
		ext->name = "VDev 320x200 dblhigh paged video driver";

		// add our video driver to the list
		return !AddVideoDriver(&ourVideoDriver);

	case	EC_CLEANUP:
		// remove our video driver from the list
		RemVideoDriver(&ourVideoDriver);
		return 0;

	default:
		return -1;	// whatever it was, we didn't do it
	};
}


boolean	ovid_init(enum VD_InitCode vdic,long x,long y) {
	void *realvm;

	switch(vdic) {
	case VD_STARTUP:
		*_vds=&ovds;
		vid_initBitMap(_vidbm,(uchar *)0xA0000,0,0,320,200,320,(*_picbm)->pal);
	case VD_REINIT:
	case VD_TO13:
		vid_setmode(0x13);
		outportb(0x3d4,0x09);
		outportb(0x3d5,(inportb(0x3d5) & 0xE0)+3);
		vid_initBitMap(_realvidbm,(uchar *)0xA0000,0,0,320,200,320,(*_picbm)->pal);
		vid_tap_reset(_vidbm);
		_vidbm->syncrect=FALSE;
		break;
	case VD_FROM13:
	case VD_CLOSEDOWN:
		outportb(0x3d4,0x09);
		outportb(0x3d5,(inportb(0x3d5)&0xE0)+1);
		break;
	case VD_GOBUFFER:
	case VD_GODIRECT:
		break;
	case VD_RESIZE:
		if (x!=320 || y!=200) return FALSE;
		break;
	};
	return TRUE;
}

void	ovid_rbitblt(coord x1,coord y1,coord xs,coord ys,uchar *mem,long dmod,boolean no_offset) {
	ulong offs=x1+y1*320;

	if (!no_offset) mem+=offs;

	rectcopy((uchar *)0xA0000+offs,320,mem,dmod,xs,ys);
};

void	ovid_rhline(coord x,coord y,uchar *mem,coord xs) {
	qmove((uchar*)0xA0000+x+320*y,mem,xs);
}

void	ovid_rhlineread(coord x,coord y,uchar *mem,coord xs) {
	qmove(mem,(uchar*)0xA0000+x+320*y,xs);
}

/* WARNING: Pointer functions are called from INTERRUPT! */

#pragma off (check_stack)

void	scroll_to(ushort y) {
	scroll_y=y;
	y*=80;

	outportb(0x3d4,0x0c);	// high
	outportb(0x3d5,y>>8);
	outportb(0x3d4,0x0d);	// low
	outportb(0x3d5,y&255);
}

void	ovid_drawptr(struct MousePointerInfo *mpi,coord x,coord y) {
	uchar *vm=(uchar *)0xA0000;
	uchar *dm1=mpi->image;
	uchar *dm2=mpi->backup;
	uchar c;
	ulong xh=mpi->s_width,yh=mpi->s_height,xht;
	ushort vmod,dmod;

	// check to see if we need to scroll the VGA

	if (mpi->cur_y < scroll_y)
		scroll_to(mpi->cur_y);
	else if (mpi->cur_y >= scroll_y+100)
		scroll_to(mpi->cur_y-99);


	if (x>=0)	vm+=x;
	if (y>=0)	vm+=(y * 320);

	mpi->vdp=vm;

	if (y<0) {
		yh+=y;
		dm1-=y*mpi->s_width;
	} else if (y>(200-mpi->s_height)) yh=200-y;

	if (x<0) {
		xh+=x;
		dm1-=x;
	} else if (x>(320-mpi->s_width)) xh=320-x;

	vmod=320-xh;
	dmod=mpi->s_width-xh;

	mpi->width=xh;
	mpi->height=yh;

	while(yh--) {
		for(xht=xh;xht;xht--) {
			*dm2++=*vm;
			if (c=*dm1++) *vm++=(c-1)?mpi->high:mpi->low; else vm++;
		};
		vm  += vmod;
		dm1 += dmod;
	};
}

void	ovid_eraseptr(struct MousePointerInfo *mpi) {
	uchar *vm=mpi->vdp;
	uchar *dm=mpi->backup;
	ushort x=mpi->width;
	ushort y=mpi->height;

	while(y--) {
		dm=qmovebes(vm,dm,x);
		vm+=320;
	};

}


