#include "vgadecod.h"

void	Decode_VGA(struct VGACapture *vc, struct VGADecode *vd) {
	uchar	overflow;

	// we have re-used the name <vd> as VGADecode * here
	// vd is *not* used to refer to video driver

	overflow = vc->crtc[7];

	vd->horiz_total		= vc->crtc[0] + 5;
	vd->horiz_displayed	= vc->crtc[1] + 1;
	vd->horiz_blank_start	= vc->crtc[2];
	vd->horiz_blank_end	= (vc->crtc[2]&0xe0)+(vc->crtc[3]&0x1f);
	if (vd->horiz_blank_end <= vd->horiz_blank_start) vd->horiz_blank_end += 0x20;

	vd->horiz_retrace_start	= vc->crtc[4];
	vd->horiz_retrace_end	= (vc->crtc[4]&0xe0)+(vc->crtc[5]&0x1f);
	if (vd->horiz_retrace_end <= vd->horiz_retrace_start) vd->horiz_retrace_end += 0x20;

	vd->vert_total		= vc->crtc[6]	+ (overflow&VGA_OVERFLOW_VTOTAL8 ? 256 : 0)
						+ (overflow&VGA_OVERFLOW_VTOTAL9 ? 512 : 0);
	vd->vert_displayed	= vc->crtc[18]	+ (overflow&VGA_OVERFLOW_VDISPENA8 ? 256 : 0)
						+ (overflow&VGA_OVERFLOW_VDISPENA9 ? 512 : 0) + 1;
	vd->vert_blank_start	= vc->crtc[21]	+ (overflow&VGA_OVERFLOW_VBLSTART8 ? 256 : 0)
						+ (vc->crtc[9]&0x20		   ? 512 : 0);
	vd->vert_blank_end	= (vd->vert_blank_start & 0xfffffff0) + (vc->crtc[22]&0x0f);

	if (vd->vert_blank_end <= vd->vert_blank_start) vd->vert_blank_end += 0x10;

	vd->vert_retrace_start	= vc->crtc[16]	+ (overflow&VGA_OVERFLOW_VRETRACE8 ? 256 : 0)
						+ (overflow&VGA_OVERFLOW_VRETRACE9 ? 512 : 0);
	vd->vert_retrace_end	= (vd->vert_retrace_start & 0xfffffff0) + (vc->crtc[17]&0x0f);
	if (vd->vert_retrace_end <= vd->vert_retrace_start) vd->vert_retrace_end += 0x10;

	vd->row_count		= (vc->crtc[9] & 0x1f)+1;
	vd->row_fine_scroll	= vc->crtc[8] & 0x1f;
	vd->row_underline	= vc->crtc[20] & 0x1f;
	vd->start_address	= vc->crtc[13] + ((ulong)vc->crtc[12]<<8);
	vd->cursor_address	= vc->crtc[15] + ((ulong)vc->crtc[14]<<8);
	vd->cursor_start	= vc->crtc[10] & 0x1f;
	vd->cursor_end		= vc->crtc[11] & 0x1f;
	vd->logical_width	= vc->crtc[19];
	vd->line_compare	= vc->crtc[24] + (overflow&VGA_OVERFLOW_LINECOMP8 ? 256 : 0)
					       + (vc->crtc[9]&0x40 ? 512 : 0);

	switch(vc->graph[6] & 0x0c) {
	case 0x00:	vd->aperture_start = (void *)0xA0000;	vd->aperture_size = 0x20000;	break;
	case 0x04:	vd->aperture_start = (void *)0xA0000;	vd->aperture_size = 0x10000;	break;
	case 0x08:	vd->aperture_start = (void *)0xB0000;	vd->aperture_size = 0x08000;	break;
	case 0x0c:	vd->aperture_start = (void *)0xB8000;	vd->aperture_size = 0x08000;	break;
	}

	vd->aperture_io		= (vc->miscout & 0x01) ? 0x3d0 : 0x3b0;

	vd->pix_mult		= ((vc->graph[5] & 0x60)==0x40) ? 4 : 8;

	switch(vc->seq[4] & 0x0c) {
	case 0x00:	vd->memory_mode = VGA_MEMORY_ODD_EVEN; break;
	case 0x04:	if ((vc->graph[5] & 0x60)==0x40)
				vd->memory_mode = VGA_MEMORY_MODEX;
			else
				vd->memory_mode = VGA_MEMORY_PLANAR;
			break;
	case 0x08:
	case 0x0c:	vd->memory_mode = VGA_MEMORY_CHAINED; break;
	}

	vd->display_mode = ((vc->miscout>>4)&0x0c) | ((vc->miscout>>2)&3);

	vd->write_mode = vc->graph[5] & 3;
}

